import configparser
import time
import gym
from gym import spaces
import numpy as np
from pandas import DataFrame
import tensorflow as tf
import threading
import os
from CMap2D import flatten_contours, render_contours_in_lidar, CMap2D, CSimAgent, fast_2f_norm
from pose2d import apply_tf_to_vel, inverse_pose2d, apply_tf_to_pose
from pkg_resources import resource_filename

import crowd_sim  # adds CrowdSim-v0 to gym  # noqa
from crowd_sim.envs.crowd_sim import CrowdSim  # reference to env code  # noqa
from crowd_sim.envs.utils.robot import Robot
from crowd_nav.policy.network_om import SDOADRL
from crowd_sim.envs.utils.action import ActionXYRot
from crowd_sim.envs.utils.info import Collision, CollisionOtherAgent, ReachGoal

PROGRESS_WEIGHT = 0.001

class SDOADRLDummyPolicy(object):
    """ the minimum viable version of a tensorflow-less SDOADRLDummyPolicy which is
    still compatible with soadrl env """
    name = 'SDOADRL'

    def configure(self, config):
        self.FOV_min_angle = config.getfloat(
            'map', 'angle_min') * np.pi % (2 * np.pi)
        self.FOV_max_angle = config.getfloat(
            'map', 'angle_max') * np.pi % (2 * np.pi)

    def human_state_in_FOV(self, self_state, human_state):
        rot = np.arctan2(
            human_state.py - self_state.py,
            human_state.px - self_state.px)
        angle = (rot - self_state.theta) % (2 * np.pi)
        return (angle > self.FOV_min_angle
                or angle < self.FOV_max_angle
                or self.FOV_min_angle == self.FOV_max_angle)

class NavRepTrainEnv(gym.Env):
    """ This class wraps the SOADRL env to make a RL ready simplified environment,
    compatible with navrep.

    Action space:
        (x velocity, y velocity, theta velocity), all in [m/s]
    Observation space:
        (scan, robotstate)
        scan:
            360 degree lidar scan, 1080 rays,
            with min angle of 0 (straight forwards), max angle of 2pi. [m]
        robotstate
            (goal x [m], goal y [m], vx [m/s], vy [m/s], vth [rad/s]) - all in robot frame
    """
    metadata = {'render.modes': ['human', 'rings']}

    def __init__(self, scenario='test', silent=False, legacy_mode=False, adaptive=True, lidar_legs=True,
                 collect_statistics=True):
        # gym env definition
        super(NavRepTrainEnv, self).__init__()
        self.action_space = spaces.Box(low=-1, high=1, shape=(3,), dtype=np.float32)
        self.observation_space = spaces.Tuple((
            spaces.Box(low=-np.inf, high=np.inf, shape=(1080,), dtype=np.float32),
            spaces.Box(low=-np.inf, high=np.inf, shape=(5,), dtype=np.float32)
        ))
        # parameters
        self.lidar_legs = lidar_legs
        self.silent = silent
        self.scenario = scenario
        self.adaptive = adaptive
        self.LEGACY_MODE = legacy_mode   # used for testing the original SOADRL policy
        self.collect_statistics = collect_statistics
        # other tools
        self.viewer = None
        self.episode_statistics = None
        if self.collect_statistics:
            self.episode_statistics = DataFrame(
                columns=[
                    "total_steps",
                    "scenario",
                    "damage",
                    "steps",
                    "goal_reached",
                    "reward",
                    "num_agents",
                    "num_walls",
                    "wall_time",
                ])
        self.total_steps = 0
        self.steps_since_reset = None
        self.episode_reward = None
        # conversion variables
        self.kLidarAngleIncrement = 0.00581718236208
        self.kLidarMergedMinAngle = 0
        self.kLidarMergedMaxAngle = 6.27543783188 + self.kLidarAngleIncrement
        self.n_angles = 1080
        self.lidar_scan = None
        self.lidar_angles = None
        self.border = None
        self.converter_cmap2d = CMap2D()
        self.converter_cmap2d.set_resolution(1.)
        self.distances_travelled_in_base_frame = None
        # environment
        self._make_env(silent=self.silent)

    def _convert_obs(self):
        robot = self.soadrl_sim.robot
        # lidar obs
        lidar_pos = np.array([robot.px, robot.py, robot.theta], dtype=np.float32)
        ranges = np.ones((self.n_angles,), dtype=np.float32) * 25.
        angles = np.linspace(self.kLidarMergedMinAngle,
                             self.kLidarMergedMaxAngle-self.kLidarAngleIncrement,
                             self.n_angles) + lidar_pos[2]
        render_contours_in_lidar(ranges, angles, self.flat_contours, lidar_pos[:2])
        # agents
        other_agents = []
        for i, human in enumerate(self.soadrl_sim.humans):
            pos = np.array([human.px, human.py, human.theta], dtype=np.float32)
            dist = self.distances_travelled_in_base_frame[i].astype(np.float32)
            vel = np.array([human.vx, human.vy], dtype=np.float32)
            if self.lidar_legs:
                agent = CSimAgent(pos, dist, vel)
            else:
                agent = CSimAgent(pos, dist, vel, type_="trunk", radius=human.radius)
            other_agents.append(agent)
        # apply through converter map (res 1., origin 0,0 -> i,j == x,y)
        self.converter_cmap2d.render_agents_in_lidar(ranges, angles, other_agents, lidar_pos[:2])
        self.lidar_scan = ranges
        self.lidar_angles = angles
        # robotstate obs
        # shape (n_agents, 5 [grx, gry, vx, vy, vtheta]) - all in base frame
        baselink_in_world = np.array([robot.px, robot.py, robot.theta])
        world_in_baselink = inverse_pose2d(baselink_in_world)
        robotvel_in_world = np.array([robot.vx, robot.vy, 0])  # TODO: actual robot rot vel?
        robotvel_in_baselink = apply_tf_to_vel(robotvel_in_world, world_in_baselink)
        goal_in_world = np.array([robot.gx, robot.gy, 0])
        goal_in_baselink = apply_tf_to_pose(goal_in_world, world_in_baselink)
        robotstate_obs = np.hstack([goal_in_baselink[:2], robotvel_in_baselink])
        obs = (self.lidar_scan, robotstate_obs)
        return obs

    def _update_dist_travelled(self):
        """ update dist travel var used for animating legs """
        # for each human, get vel in base frame
        for i, human in enumerate(self.soadrl_sim.humans):
            # dig up rotational velocity from past states log
            vrot = 0.
            if len(self.soadrl_sim.states) > 1:
                vrot = (self.soadrl_sim.states[-1][1][i].theta
                        - self.soadrl_sim.states[-2][1][i].theta) / self._get_dt()
            # transform world vel to base vel
            baselink_in_world = np.array([human.px, human.py, human.theta])
            world_in_baselink = inverse_pose2d(baselink_in_world)
            vel_in_world_frame = np.array([human.vx, human.vy, vrot])
            vel_in_baselink_frame = apply_tf_to_vel(vel_in_world_frame, world_in_baselink)
            self.distances_travelled_in_base_frame[i, :] += vel_in_baselink_frame * self._get_dt()

    def _add_border_obstacle(self):
        all_agents = [self.soadrl_sim.robot] + self.soadrl_sim.humans + self.soadrl_sim.other_robots
        all_vertices = np.array(self.soadrl_sim.obstacle_vertices).reshape((-1,2))
        x_agents = [a.px for a in all_agents]
        y_agents = [a.py for a in all_agents]
        x_goals = [a.gx for a in all_agents]
        y_goals = [a.gy for a in all_agents]
        x_vertices = list(all_vertices[:,0])
        y_vertices = list(all_vertices[:,1])
        BORDER_OFFSET = 1.
        x_min = min(x_agents + x_goals + x_vertices) - BORDER_OFFSET
        x_max = max(x_agents + x_goals + x_vertices) + BORDER_OFFSET
        y_min = min(y_agents + y_goals + y_vertices) - BORDER_OFFSET
        y_max = max(y_agents + y_goals + y_vertices) + BORDER_OFFSET
        # anticlockwise -> bounding obstacle
        border_vertices = [(x_min, y_min), (x_max, y_min), (x_max, y_max), (x_min, y_max)][::-1]
        self.border = [(x_min, x_max), (y_min, y_max)]
        self.soadrl_sim.obstacle_vertices.append(border_vertices)

    def step(self, action):
        self.steps_since_reset += 1
        self.total_steps += 1
        if not self.LEGACY_MODE:
            # convert action
            # SOADRL - rotation is dtheta
            # IAN    - rotation is dtheta/dt
            dr = action[2] * self._get_dt()
            #  SOADRL - instant rot, then vel
            #  IAN    - vel, then rot
            vx_baselink = action[0] * np.cos(-dr) - action[1] * np.sin(-dr)
            vy_baselink = action[0] * np.sin(-dr) + action[1] * np.cos(-dr)
            soadrl_action = ActionXYRot(vx_baselink, vy_baselink, dr)
            # step
            _, _, reward, done, info = self.soadrl_sim.step(
                soadrl_action, compute_local_map=False, border=self.border)
            self._update_dist_travelled()
            # convert observations
            obs = self._convert_obs()
        else:
            state, local_map, reward, done, info = self.soadrl_sim.step(
                action, compute_local_map=True, border=self.border)
            self._update_dist_travelled()
            self._convert_obs()
            obs = (state, local_map)
            self.lidar_scan = local_map * self.soadrl_sim.angular_map_max_range
            self.lidar_angles = np.linspace(self.soadrl_sim.angular_map_min_angle,
                                            self.soadrl_sim.angular_map_max_angle,
                                            len(local_map)
                                            ) + self.soadrl_sim.robot.theta
        # add progress reward
        progress_reward = 0
        if not done:
            if len(self.soadrl_sim.states) > 1:
                new_s = self.soadrl_sim.states[-1][0]
                old_s = self.soadrl_sim.states[-2][0]
                new_p = np.array([new_s.px, new_s.py], dtype=np.double)
                old_p = np.array([old_s.px, old_s.py], dtype=np.double)
                goal = np.array([new_s.gx, new_s.gy], dtype=np.double)
                progress_reward = PROGRESS_WEIGHT*(fast_2f_norm(old_p - goal) - fast_2f_norm(new_p - goal))
        reward = reward + progress_reward
        reward = reward * 100.
        self.episode_reward += reward
        # adaptive difficulty
        if done and self.adaptive and self.scenario == "train":
            if isinstance(info, ReachGoal):
                self.soadrl_sim.human_num += 1
                self.soadrl_sim.num_walls += 1
                self.soadrl_sim.num_circles += 1
            else:
                self.soadrl_sim.human_num -= 1
                self.soadrl_sim.num_walls -= 1
                self.soadrl_sim.num_circles -= 1
            self.soadrl_sim.human_num = np.clip(self.soadrl_sim.human_num, 0, 5)
            self.soadrl_sim.num_walls = np.clip(self.soadrl_sim.num_walls, 0, 10)
            self.soadrl_sim.num_circles = np.clip(self.soadrl_sim.num_circles, 0, 10)
        # log data
        if done:
            if self.collect_statistics:
                self.episode_statistics.loc[len(self.episode_statistics)] = [
                    self.total_steps,
                    'navreptrain'+self.scenario,
                    100 if isinstance(info, Collision) or isinstance(info, CollisionOtherAgent) else 0,
                    self.steps_since_reset,
                    isinstance(info, ReachGoal),
                    reward,
                    self.soadrl_sim.human_num,
                    self.soadrl_sim.num_walls,
                    time.time(),
                ]
        return obs, reward, done, {'event': info}

    def reset(self):
        self.steps_since_reset = 0
        self.episode_reward = 0
        _, _ = self.soadrl_sim.reset(self.scenario, compute_local_map=False)
        random_rot = ActionXYRot(0, 0, 10.*(np.random.random()-0.5))
        self.soadrl_sim.step(random_rot, compute_local_map=False, border=self.border)
        if not self.LEGACY_MODE:
            self._add_border_obstacle()
        contours = self.soadrl_sim.obstacle_vertices
        self.flat_contours = flatten_contours(contours)
        self.distances_travelled_in_base_frame = np.zeros((len(self.soadrl_sim.humans), 3))
        obs = self._convert_obs()
        if self.LEGACY_MODE:
            state, local_map, reward, done, info = self.soadrl_sim.step(
                ActionXYRot(0, 0, 0), compute_local_map=True, border=self.border)
            obs = (state, local_map)
        return obs

    def render(self, mode='human', close=False,
               RENDER_LIDAR=True, lidar_scan_override=None, goal_override=None, save_to_file=False,
               show_score=False, robocentric=False):
        if close:
            if self.viewer is not None:
                self.viewer.close()
            return
        if self.lidar_scan is None:  # check that reset has been called
            return
        if mode == 'traj':
            self.soadrl_sim.render('traj')
        elif mode in ['human', 'rings']:
            # Window and viewport size
            WINDOW_W = 256
            WINDOW_H = 256
            VP_W = WINDOW_W
            VP_H = WINDOW_H
            from gym.envs.classic_control import rendering
            import pyglet
            from pyglet import gl
            # Create viewer
            if self.viewer is None:
                self.viewer = rendering.Viewer(WINDOW_W, WINDOW_H)
                self.transform = rendering.Transform()
                self.transform.set_scale(10, 10)
                self.transform.set_translation(128, 128)
                self.score_label = pyglet.text.Label(
                    '0000', font_size=12,
                    x=20, y=WINDOW_H*2.5/40.00, anchor_x='left', anchor_y='center',
                    color=(255,255,255,255))
#                 self.transform = rendering.Transform()
                self.currently_rendering_iteration = 0
                self.image_lock = threading.Lock()

            def make_circle(c, r, res=10):
                thetas = np.linspace(0, 2*np.pi, res+1)[:-1]
                verts = np.zeros((res, 2))
                verts[:,0] = c[0] + r * np.cos(thetas)
                verts[:,1] = c[1] + r * np.sin(thetas)
                return verts
            # Render in pyglet
            with self.image_lock:
                self.currently_rendering_iteration += 1
                self.viewer.draw_circle(r=10, color=(0.3,0.3,0.3))
                win = self.viewer.window
                win.switch_to()
                win.dispatch_events()
                win.clear()
                gl.glViewport(0, 0, VP_W, VP_H)
                # colors
                bgcolor = np.array([0.4, 0.8, 0.4])
                obstcolor = np.array([0.3, 0.3, 0.3])
                goalcolor = np.array([1., 1., 0.3])
                goallinecolor = 0.9 * bgcolor
                nosecolor = np.array([0.3, 0.3, 0.3])
                lidarcolor = np.array([1., 0., 0.])
                agentcolor = np.array([0., 1., 1.])
                # Green background
                gl.glBegin(gl.GL_QUADS)
                gl.glColor4f(bgcolor[0], bgcolor[1], bgcolor[2], 1.0)
                gl.glVertex3f(0, VP_H, 0)
                gl.glVertex3f(VP_W, VP_H, 0)
                gl.glVertex3f(VP_W, 0, 0)
                gl.glVertex3f(0, 0, 0)
                gl.glEnd()
                # Transform
                rx = self.soadrl_sim.robot.px
                ry = self.soadrl_sim.robot.py
                rth = self.soadrl_sim.robot.theta
                if robocentric:
                    # sets viewport = robocentric a.k.a T_sim_in_viewport = T_sim_in_robocentric
                    from pose2d import inverse_pose2d
                    T_sim_in_robot = inverse_pose2d(np.array([rx, ry, rth]))
                    # T_robot_in_robocentric is trans(128, 128), scale(10), rot(90deg)
                    # T_sim_in_robocentric = T_sim_in_robot * T_robot_in_robocentric
                    rot = np.pi / 2.
                    scale = 20
                    trans = (WINDOW_W / 2., WINDOW_H / 2.)
                    T_sim_in_robocentric = [
                        trans[0] + scale * (T_sim_in_robot[0] * np.cos(rot) - T_sim_in_robot[1] * np.sin(rot)),
                        trans[1] + scale * (T_sim_in_robot[0] * np.sin(rot) + T_sim_in_robot[1] * np.cos(rot)),
                        T_sim_in_robot[2] + rot,
                    ]
                    self.transform.set_translation(T_sim_in_robocentric[0], T_sim_in_robocentric[1])
                    self.transform.set_rotation(T_sim_in_robocentric[2])
                    self.transform.set_scale(scale, scale)
#                     self.transform.set_scale(20, 20)
                self.transform.enable()  # applies T_sim_in_viewport to below coords (all in sim frame)
                # Map closed obstacles ---
                for poly in self.soadrl_sim.obstacle_vertices:
                    gl.glBegin(gl.GL_LINE_LOOP)
                    gl.glColor4f(obstcolor[0], obstcolor[1], obstcolor[2], 1)
                    for vert in poly:
                        gl.glVertex3f(vert[0], vert[1], 0)
                    gl.glEnd()
                # LIDAR
                if RENDER_LIDAR:
                    px = self.soadrl_sim.robot.px
                    py = self.soadrl_sim.robot.py
                    angle = self.soadrl_sim.robot.theta
                    # LIDAR rays
                    scan = lidar_scan_override
                    if scan is None:
                        scan = self.lidar_scan
                    lidar_angles = self.lidar_angles
                    x_ray_ends = px + scan * np.cos(lidar_angles)
                    y_ray_ends = py + scan * np.sin(lidar_angles)
                    is_in_fov = np.cos(lidar_angles - angle) >= 0.78
                    for ray_idx in range(len(scan)):
                        end_x = x_ray_ends[ray_idx]
                        end_y = y_ray_ends[ray_idx]
                        gl.glBegin(gl.GL_LINE_LOOP)
                        if is_in_fov[ray_idx]:
                            gl.glColor4f(1., 1., 0., 0.1)
                        else:
                            gl.glColor4f(lidarcolor[0], lidarcolor[1], lidarcolor[2], 0.1)
                        gl.glVertex3f(px, py, 0)
                        gl.glVertex3f(end_x, end_y, 0)
                        gl.glEnd()
                # Agent body
                for n, agent in enumerate([self.soadrl_sim.robot] + self.soadrl_sim.humans):
                    px = agent.px
                    py = agent.py
                    angle = agent.theta
                    r = agent.radius
                    # Agent as Circle
                    poly = make_circle((px, py), r)
                    gl.glBegin(gl.GL_POLYGON)
                    if n == 0:
                        color = np.array([1., 1., 1.])
                    else:
                        color = agentcolor
                    gl.glColor4f(color[0], color[1], color[2], 1)
                    for vert in poly:
                        gl.glVertex3f(vert[0], vert[1], 0)
                    gl.glEnd()
                    # Direction triangle
                    xnose = px + r * np.cos(angle)
                    ynose = py + r * np.sin(angle)
                    xright = px + 0.3 * r * -np.sin(angle)
                    yright = py + 0.3 * r * np.cos(angle)
                    xleft = px - 0.3 * r * -np.sin(angle)
                    yleft = py - 0.3 * r * np.cos(angle)
                    gl.glBegin(gl.GL_TRIANGLES)
                    gl.glColor4f(nosecolor[0], nosecolor[1], nosecolor[2], 1)
                    gl.glVertex3f(xnose, ynose, 0)
                    gl.glVertex3f(xright, yright, 0)
                    gl.glVertex3f(xleft, yleft, 0)
                    gl.glEnd()
                # Goal
                xgoal = self.soadrl_sim.robot.gx
                ygoal = self.soadrl_sim.robot.gy
                r = self.soadrl_sim.robot.radius
                if goal_override is not None:
                    xgoal, ygoal = goal_override
                # Goal markers
                gl.glBegin(gl.GL_TRIANGLES)
                gl.glColor4f(goalcolor[0], goalcolor[1], goalcolor[2], 1)
                triangle = make_circle((xgoal, ygoal), r, res=3)
                for vert in triangle:
                    gl.glVertex3f(vert[0], vert[1], 0)
                gl.glEnd()
                # Goal line
                gl.glBegin(gl.GL_LINE_LOOP)
                gl.glColor4f(goallinecolor[0], goallinecolor[1], goallinecolor[2], 1)
                gl.glVertex3f(rx, ry, 0)
                gl.glVertex3f(xgoal, ygoal, 0)
                gl.glEnd()
                # --
                self.transform.disable()
                # Text
                self.score_label.text = ""
                if show_score:
                    self.score_label.text = "R {}".format(self.episode_reward)
                self.score_label.draw()
                win.flip()
                if save_to_file:
                    pyglet.image.get_buffer_manager().get_color_buffer().save(
                        "/tmp/navreptrainenv{:05}.png".format(self.total_steps))
                return self.viewer.isopen
        else:
            raise NotImplementedError

    def close(self):
        self.render(close=True)

    def _make_env(self, silent=False):
        # Create env
        config_dir = resource_filename('crowd_nav', 'config')
        config_file = os.path.join(config_dir, 'test_soadrl_static.config')
        config_file = os.path.expanduser(config_file)
        config = configparser.RawConfigParser()
        config.read(config_file)

        env = gym.make('CrowdSim-v0')
        env.configure(config, silent=silent)
        robot = Robot(config, 'humans')
        env.set_robot(robot)

        policy = SDOADRLDummyPolicy()
        policy.configure(config)
        if self.LEGACY_MODE:
            sess = tf.Session()
            policy = SDOADRL()
            policy.configure(sess, 'global', config)
            policy.set_phase('test')
            policy.load_model(os.path.expanduser('~/soadrl/Final_models/angular_map_full_FOV/rl_model'))

        env.robot.set_policy(policy)
        if not silent:
            env.robot.print_info()

        self.soadrl_sim = env

    def _get_viewer(self):
        return self.viewer

    def _get_dt(self):
        return self.soadrl_sim.time_step


if __name__ == "__main__":
    from navrep.tools.envplayer import EnvPlayer

    env = NavRepTrainEnv()
    player = EnvPlayer(env)



import os
from matplotlib import pyplot as plt

from navrep.tools.commonargs import parse_plotting_args
from plot_gym_training_progress import plot_per_scenario_training_progress

SCENARIO_AXES = {
    "navreptrain" : (0, 0),
    "navrepval" : (1, 0),
}
ERROR_IF_MISSING = False

if __name__ == "__main__":
    args, _ = parse_plotting_args()

    if args.x_axis is None:
        args.x_axis = "train_steps"

    while True:
        basedir = args.logdir
        if basedir is None:
            basedir = "~/best_navrep"
        basedir = os.path.expanduser(basedir)
        best_navrep_dirs = [os.path.join(basedir, o) for o in os.listdir(basedir)
                            if os.path.isdir(os.path.join(basedir,o))]
        plot_per_scenario_training_progress(best_navrep_dirs, SCENARIO_AXES, ERROR_IF_MISSING,
                                            x_axis=args.x_axis, paper_ready=True)
        plt.xlim([-5, 61])
        plt.pause(60)



import numpy as np
from gym import spaces

from navrep.tools.rings import generate_rings
from navrep.envs.navreptrainenv import NavRepTrainEnv
from navrep.envs.ianenv import IANEnv

_L = 1080  # lidar size
_RS = 5  # robotstate size
_64 = 64  # ring size

class FlatLidarAndStateEncoder(object):
    """ Generic class to encode the observations of an environment into a single 1d vector """
    def __init__(self):
        self._N = _L + _RS
        self.observation_space = spaces.Box(low=-np.inf, high=np.inf,
                                            shape=(self._N,1), dtype=np.float32)

    def reset(self):
        pass

    def close(self):
        pass

    def _encode_obs(self, obs, action):
        lidar, state = obs
        e2e_obs = np.concatenate([lidar, state]).reshape(self._N,1)
        return e2e_obs

class RingsLidarAndStateEncoder(object):
    """ Generic class to encode the observations of an environment into a rings 2d image """
    def __init__(self):
        self._N = _64*_64 + _RS
        self.observation_space = spaces.Box(low=-np.inf, high=np.inf,
                                            shape=(self._N,1), dtype=np.float32)
        self.rings_def = generate_rings(_64, _64)

    def reset(self):
        pass

    def close(self):
        pass

    def _encode_obs(self, obs, action):
        lidar, state = obs
        rings = (
            self.rings_def["lidar_to_rings"](lidar[None, :]).astype(float)
            / self.rings_def["rings_to_bool"]
        )
        e2e_obs = np.concatenate([rings.reshape(_64*_64), state]).reshape(self._N,1)
        return e2e_obs

class E2E1DNavRepEnv(NavRepTrainEnv):
    """ takes a (2) action as input
    outputs encoded obs (1085) """
    def __init__(self, *args, **kwargs):
        self.encoder = FlatLidarAndStateEncoder()
        super(E2E1DNavRepEnv, self).__init__(*args, **kwargs)
        self.action_space = spaces.Box(low=-1, high=1, shape=(2,), dtype=np.float32)
        self.observation_space = self.encoder.observation_space

    def step(self, action):
        action = np.array([action[0], action[1], 0.])  # no rotation
        obs, reward, done, info = super(E2E1DNavRepEnv, self).step(action)
        h = self.encoder._encode_obs(obs, action)
        return h, reward, done, info

    def reset(self):
        self.encoder.reset()
        obs = super(E2E1DNavRepEnv, self).reset()
        h = self.encoder._encode_obs(obs, np.array([0,0,0]))
        return h

class E2ENavRepEnv(NavRepTrainEnv):
    """ takes a (2) action as input
    outputs encoded obs (1085) """
    def __init__(self, *args, **kwargs):
        self.encoder = RingsLidarAndStateEncoder()
        super(E2ENavRepEnv, self).__init__(*args, **kwargs)
        self.action_space = spaces.Box(low=-1, high=1, shape=(2,), dtype=np.float32)
        self.observation_space = self.encoder.observation_space

    def step(self, action):
        action = np.array([action[0], action[1], 0.])  # no rotation
        obs, reward, done, info = super(E2ENavRepEnv, self).step(action)
        h = self.encoder._encode_obs(obs, action)
        return h, reward, done, info

    def reset(self):
        self.encoder.reset()
        obs = super(E2ENavRepEnv, self).reset()
        h = self.encoder._encode_obs(obs, np.array([0,0,0]))
        return h

class E2E1DIANEnv(IANEnv):
    """ takes a (2) action as input
    outputs encoded obs (1085) """
    def __init__(self, *args, **kwargs):
        self.encoder = FlatLidarAndStateEncoder()
        super(E2E1DIANEnv, self).__init__(*args, **kwargs)
        self.action_space = spaces.Box(low=-1, high=1, shape=(2,), dtype=np.float32)
        self.observation_space = self.encoder.observation_space

    def step(self, action):
        action = np.array([action[0], action[1], 0.])  # no rotation
        obs, reward, done, info = super(E2E1DIANEnv, self).step(action)
        h = self.encoder._encode_obs(obs, action)
        return h, reward, done, info

    def reset(self):
        self.encoder.reset()
        obs = super(E2E1DIANEnv, self).reset()
        h = self.encoder._encode_obs(obs, np.array([0,0,0]))
        return h

class E2EIANEnv(IANEnv):
    """ takes a (2) action as input
    outputs encoded obs (1085) """
    def __init__(self, *args, **kwargs):
        self.encoder = RingsLidarAndStateEncoder()
        super(E2EIANEnv, self).__init__(*args, **kwargs)
        self.action_space = spaces.Box(low=-1, high=1, shape=(2,), dtype=np.float32)
        self.observation_space = self.encoder.observation_space

    def step(self, action):
        action = np.array([action[0], action[1], 0.])  # no rotation
        obs, reward, done, info = super(E2EIANEnv, self).step(action)
        h = self.encoder._encode_obs(obs, action)
        return h, reward, done, info

    def reset(self):
        self.encoder.reset()
        obs = super(E2EIANEnv, self).reset()
        h = self.encoder._encode_obs(obs, np.array([0,0,0]))
        return h



# This Python file uses the following encoding: utf-8
"""autogenerated by genpy from frame_msgs/TrackedPersons2d.msg. Do not edit."""
import sys
python3 = True if sys.hexversion > 0x03000000 else False
import genpy
import struct

import frame_msgs.msg
import std_msgs.msg

class TrackedPersons2d(genpy.Message):
  _md5sum = "972b7d693ce31c3b18f092f43387621c"
  _type = "frame_msgs/TrackedPersons2d"
  _has_header = True #flag to mark the presence of a Header object
  _full_text = """# Message with all 2d bbox in image of currently tracked persons 
#

Header                header      # Header containing timestamp etc. of this message
TrackedPerson2d[]     boxes       # All persons that are currently being tracked (2d image bbox)

================================================================================
MSG: std_msgs/Header
# Standard metadata for higher-level stamped data types.
# This is generally used to communicate timestamped data 
# in a particular coordinate frame.
# 
# sequence ID: consecutively increasing ID 
uint32 seq
#Two-integer timestamp that is expressed as:
# * stamp.sec: seconds (stamp_secs) since epoch (in Python the variable is called 'secs')
# * stamp.nsec: nanoseconds since stamp_secs (in Python the variable is called 'nsecs')
# time-handling sugar is provided by the client library
time stamp
#Frame this data is associated with
string frame_id

================================================================================
MSG: frame_msgs/TrackedPerson2d
# Message defining a 2d image bbox of a tracked person
#

uint64      track_id        # unique identifier of the target, consistent over time
float32     person_height   # 3d height of person in m
int32       x               # top left corner x of 2d image bbox
int32       y               # top left corner y of 2d image bbox
uint32      w               # width of 2d image bbox
uint32      h               # height of 2d image bbox
float32     depth           # distance from the camera in m
"""
  __slots__ = ['header','boxes']
  _slot_types = ['std_msgs/Header','frame_msgs/TrackedPerson2d[]']

  def __init__(self, *args, **kwds):
    """
    Constructor. Any message fields that are implicitly/explicitly
    set to None will be assigned a default value. The recommend
    use is keyword arguments as this is more robust to future message
    changes.  You cannot mix in-order arguments and keyword arguments.

    The available fields are:
       header,boxes

    :param args: complete set of field values, in .msg order
    :param kwds: use keyword arguments corresponding to message field names
    to set specific fields.
    """
    if args or kwds:
      super(TrackedPersons2d, self).__init__(*args, **kwds)
      #message fields cannot be None, assign default values for those that are
      if self.header is None:
        self.header = std_msgs.msg.Header()
      if self.boxes is None:
        self.boxes = []
    else:
      self.header = std_msgs.msg.Header()
      self.boxes = []

  def _get_types(self):
    """
    internal API method
    """
    return self._slot_types

  def serialize(self, buff):
    """
    serialize message into buffer
    :param buff: buffer, ``StringIO``
    """
    try:
      _x = self
      buff.write(_get_struct_3I().pack(_x.header.seq, _x.header.stamp.secs, _x.header.stamp.nsecs))
      _x = self.header.frame_id
      length = len(_x)
      if python3 or type(_x) == unicode:
        _x = _x.encode('utf-8')
        length = len(_x)
      buff.write(struct.pack('<I%ss'%length, length, _x))
      length = len(self.boxes)
      buff.write(_struct_I.pack(length))
      for val1 in self.boxes:
        _x = val1
        buff.write(_get_struct_Qf2i2If().pack(_x.track_id, _x.person_height, _x.x, _x.y, _x.w, _x.h, _x.depth))
    except struct.error as se: self._check_types(struct.error("%s: '%s' when writing '%s'" % (type(se), str(se), str(locals().get('_x', self)))))
    except TypeError as te: self._check_types(ValueError("%s: '%s' when writing '%s'" % (type(te), str(te), str(locals().get('_x', self)))))

  def deserialize(self, str):
    """
    unpack serialized message in str into this message instance
    :param str: byte array of serialized message, ``str``
    """
    try:
      if self.header is None:
        self.header = std_msgs.msg.Header()
      if self.boxes is None:
        self.boxes = None
      end = 0
      _x = self
      start = end
      end += 12
      (_x.header.seq, _x.header.stamp.secs, _x.header.stamp.nsecs,) = _get_struct_3I().unpack(str[start:end])
      start = end
      end += 4
      (length,) = _struct_I.unpack(str[start:end])
      start = end
      end += length
      if python3:
        self.header.frame_id = str[start:end].decode('utf-8')
      else:
        self.header.frame_id = str[start:end]
      start = end
      end += 4
      (length,) = _struct_I.unpack(str[start:end])
      self.boxes = []
      for i in range(0, length):
        val1 = frame_msgs.msg.TrackedPerson2d()
        _x = val1
        start = end
        end += 32
        (_x.track_id, _x.person_height, _x.x, _x.y, _x.w, _x.h, _x.depth,) = _get_struct_Qf2i2If().unpack(str[start:end])
        self.boxes.append(val1)
      return self
    except struct.error as e:
      raise genpy.DeserializationError(e) #most likely buffer underfill


  def serialize_numpy(self, buff, numpy):
    """
    serialize message with numpy array types into buffer
    :param buff: buffer, ``StringIO``
    :param numpy: numpy python module
    """
    try:
      _x = self
      buff.write(_get_struct_3I().pack(_x.header.seq, _x.header.stamp.secs, _x.header.stamp.nsecs))
      _x = self.header.frame_id
      length = len(_x)
      if python3 or type(_x) == unicode:
        _x = _x.encode('utf-8')
        length = len(_x)
      buff.write(struct.pack('<I%ss'%length, length, _x))
      length = len(self.boxes)
      buff.write(_struct_I.pack(length))
      for val1 in self.boxes:
        _x = val1
        buff.write(_get_struct_Qf2i2If().pack(_x.track_id, _x.person_height, _x.x, _x.y, _x.w, _x.h, _x.depth))
    except struct.error as se: self._check_types(struct.error("%s: '%s' when writing '%s'" % (type(se), str(se), str(locals().get('_x', self)))))
    except TypeError as te: self._check_types(ValueError("%s: '%s' when writing '%s'" % (type(te), str(te), str(locals().get('_x', self)))))

  def deserialize_numpy(self, str, numpy):
    """
    unpack serialized message in str into this message instance using numpy for array types
    :param str: byte array of serialized message, ``str``
    :param numpy: numpy python module
    """
    try:
      if self.header is None:
        self.header = std_msgs.msg.Header()
      if self.boxes is None:
        self.boxes = None
      end = 0
      _x = self
      start = end
      end += 12
      (_x.header.seq, _x.header.stamp.secs, _x.header.stamp.nsecs,) = _get_struct_3I().unpack(str[start:end])
      start = end
      end += 4
      (length,) = _struct_I.unpack(str[start:end])
      start = end
      end += length
      if python3:
        self.header.frame_id = str[start:end].decode('utf-8')
      else:
        self.header.frame_id = str[start:end]
      start = end
      end += 4
      (length,) = _struct_I.unpack(str[start:end])
      self.boxes = []
      for i in range(0, length):
        val1 = frame_msgs.msg.TrackedPerson2d()
        _x = val1
        start = end
        end += 32
        (_x.track_id, _x.person_height, _x.x, _x.y, _x.w, _x.h, _x.depth,) = _get_struct_Qf2i2If().unpack(str[start:end])
        self.boxes.append(val1)
      return self
    except struct.error as e:
      raise genpy.DeserializationError(e) #most likely buffer underfill

_struct_I = genpy.struct_I
def _get_struct_I():
    global _struct_I
    return _struct_I
_struct_3I = None
def _get_struct_3I():
    global _struct_3I
    if _struct_3I is None:
        _struct_3I = struct.Struct("<3I")
    return _struct_3I
_struct_Qf2i2If = None
def _get_struct_Qf2i2If():
    global _struct_Qf2i2If
    if _struct_Qf2i2If is None:
        _struct_Qf2i2If = struct.Struct("<Qf2i2If")
    return _struct_Qf2i2If



from ._DetectedPerson import *
from ._DetectedPersons import *
from ._PersonTrajectories import *
from ._PersonTrajectory import *
from ._PersonTrajectoryEntry import *
from ._TrackedPerson import *
from ._TrackedPerson2d import *
from ._TrackedPersons import *
from ._TrackedPersons2d import *



from __future__ import print_function
import numpy as np
import os

from navrep.models.siren1d import SIREN1D, reset_graph
from navrep.tools.data_extraction import archive_to_lidar_dataset

DEBUG_PLOTTING = True

# Parameters for training
batch_size = 100
NUM_EPOCH = 10
DATA_DIR = "record"
HOME = os.path.expanduser("~")
MAX_LIDAR_DIST = 25.0

model_save_dir = HOME + "/navrep/models/V"
model_save_path = os.path.join(model_save_dir, "siren1d.json")
if not os.path.exists(model_save_dir):
    os.makedirs(model_save_dir)

# create network
reset_graph()
siren = SIREN1D(batch_size=batch_size, is_training=True, reuse=False)

# create training dataset
dataset = archive_to_lidar_dataset(limit=1)
if len(dataset) == 0:
    raise ValueError("no scans found, exiting")
print(len(dataset), "scans in dataset.")
n_scans = len(dataset)

all_obs = np.clip(dataset.astype(np.float) / MAX_LIDAR_DIST, 0.0, MAX_LIDAR_DIST)
train_o = all_obs[:,None,:].reshape(n_scans, 1, 1080, 1)
train_x = np.arange(1080)[None, :].reshape(1, 1080, 1)  # TODO: divide by 1080
train_y = all_obs.reshape((n_scans, 1080, 1))

# batch = sample from (n_scans, n_points)
all_indices = np.array(np.where(all_obs)).T

# split into batches:
num_batches = int(np.floor(len(all_indices) / batch_size))

# train loop:
print("train", "step", "loss", "recon_loss")
for epoch in range(NUM_EPOCH):
    np.random.shuffle(all_indices)
    for idx in range(num_batches):
        batch_indices = all_indices[idx * batch_size : (idx + 1) * batch_size]
        batch_o = train_o[batch_indices[:,0], 0]
        batch_x = train_x[0, batch_indices[:,1]]
        batch_y = train_y[batch_indices[:,0], batch_indices[:,1]]

        feed = {
            siren.x: batch_x,
            siren.o: batch_o,
            siren.y_true: batch_y,
        }

        (train_loss, r_loss, train_step, y_pred, _) = siren.sess.run(
            [siren.loss, siren.r_loss, siren.global_step, siren.y, siren.train_op], feed
        )

        print("{:>4.1f}%".format(100. * (train_step % 500) / 500,), end="\r")
        if (train_step + 1) % 500 == 0:
            print("step", (train_step + 1), train_loss, r_loss)
        if (train_step + 1) % 5000 == 0:
            siren.save_json(model_save_path)
            if DEBUG_PLOTTING:
                obs = batch_o[0]
                full_y = np.zeros((1080,))
                for i in range(0, 1000, batch_size):
                    print("{:>4.1f}%".format(i / 10.,), end="\r")
                    batch_y = obs[i:i+batch_size]
                    batch_x = np.arange(i, i+100).reshape((batch_size, 1))
                    batch_o = np.repeat(obs[None,:], batch_size, axis=0)
                    feed = {
                        siren.x: batch_x,
                        siren.o: batch_o,
                        siren.y_true: batch_y,
                    }
                    y_pred = siren.sess.run(siren.y, feed)
                    full_y[i:i+batch_size] = np.squeeze(y_pred)
                from matplotlib import pyplot as plt

                plt.ion()
                plt.cla()
                plt.plot(obs[:,0])
                plt.plot(full_y)
                plt.title("epoch #{}".format(epoch))
                plt.pause(0.01)

print("Done.")

# finished, final model:
siren.save_json(model_save_path)



# -*- coding: utf-8 -*-
# generated from catkin/cmake/template/__init__.py.in
# keep symbol table as clean as possible by deleting all unnecessary symbols

from os import path as os_path
from pkgutil import extend_path
from sys import path as sys_path
import pkg_resources

navrep_dir = pkg_resources.resource_filename('navrep', '')
__extended_path = [os_path.join(navrep_dir, '../external/tf_bag/src')]
for p in reversed(__extended_path):
    sys_path.insert(0, p)
    del p
del sys_path

__path__ = extend_path(__path__, __name__)
del extend_path

__execfiles = []
for p in __extended_path:
    src_init_file = os_path.join(p, __name__ + '.py')
    if os_path.isfile(src_init_file):
        __execfiles.append(src_init_file)
    else:
        src_init_file = os_path.join(p, __name__, '__init__.py')
        if os_path.isfile(src_init_file):
            __execfiles.append(src_init_file)
    del src_init_file
    del p
del os_path
del __extended_path

for __execfile in __execfiles:
    with open(__execfile, 'r') as __fh:
        exec(__fh.read())
    del __fh
    del __execfile
del __execfiles



import numpy as np
import logging
import torch
import torch.nn as nn
from torch.nn import functional as F

from navrep.models.torchvae import VAE1D
from navrep.models.gpt import Block

_A = 3
_G = 2
_H = 64
_Z = _H

logger = logging.getLogger(__name__)

STATE_NORM_FACTOR = 25.  # maximum typical goal distance, meters

class GPT1D(nn.Module):
    def __init__(self, config, gpu=True):
        super().__init__()

        # input embedding stem
        self.convVAE = VAE1D(z_dim=config.n_embd, gpu=gpu)
        self.action_emb = nn.Linear(config.n_action, config.n_embd)
        self.pos_emb = nn.Parameter(torch.zeros(1, config.block_size, config.n_embd))
        self.state_emb = nn.Linear(config.n_states, config.n_embd)
        self.drop = nn.Dropout(config.embd_pdrop)
        # transformer
        self.blocks = nn.Sequential(*[Block(config) for _ in range(config.n_layer)])
        # decoder head
        self.ln_f = nn.LayerNorm(config.n_embd)
        self.z_head = nn.Linear(config.n_embd, config.n_embd)
        self.state_head = nn.Linear(config.n_embd, config.n_states)

        self.block_size = config.block_size
        self.n_embd = config.n_embd
        self.apply(self._init_weights)

        self.gpu = gpu

        logger.info(
            "number of parameters: %e", sum(p.numel() for p in self.parameters())
        )

    def _init_weights(self, module):
        if isinstance(module, (nn.Linear, nn.Embedding)):
            module.weight.data.normal_(mean=0.0, std=0.02)
            if isinstance(module, nn.Linear) and module.bias is not None:
                module.bias.data.zero_()
        elif isinstance(module, nn.LayerNorm):
            module.bias.data.zero_()
            module.weight.data.fill_(1.0)

    def get_block_size(self):
        return self.block_size

    def forward(self, lidar, state, action, dones, targets=None, h=None):
        """
        lidar: (batch, sequence, CH, L) [0, 1]
        action: (batch, sequence, A) [-inf, inf]
        state: (batch, sequence, S) [-inf, inf]
        dones:  (batch, sequence,) {0, 1}
        targets: None or (lidar_targets, state_targets)
            lidar_targets: same shape as lidar
            state_targets: same shape as state

        OUTPUTS
        lidar_pred: same shape as lidar
        state_pred: same shape as state
        loss: torch loss
        """
        b, t, CH, L = lidar.size()
        _, _, A = action.size()
        _, _, S = state.size()
        assert t <= self.block_size, "Cannot forward, model block size is exhausted."

        # encode embedding with vae
        z, mu, logvar = self.convVAE.encode(lidar.view(b * t, CH, L))  # each image maps to a vector
        token_embeddings = z.view(b, t, self.n_embd)
        state_embeddings = self.state_emb(state.view(b * t, S)).view(b, t, self.n_embd)
        action_embeddings = self.action_emb(action.view(b * t, A)).view(b, t, self.n_embd)
        position_embeddings = self.pos_emb[:, :t, :]  # each position maps to a (learnable) vector
        # forward the GPT model
        x = self.drop(token_embeddings + position_embeddings + action_embeddings + state_embeddings)
        x = self.blocks(x)
        x = self.ln_f(x)
        # store worldmodel embedding
        if h is not None:
            h[0] = x
        # decode embedding with vae
        z_pred = self.z_head(x.view(b * t, self.n_embd))
        lidar_rec = self.convVAE.decode(z).view(b, t, CH, L)
        lidar_pred = self.convVAE.decode(z_pred).view(b, t, CH, L)
        state_pred = self.state_head(x.view(b * t, self.n_embd)).view(b, t, S)

        # if we are given some desired targets also calculate the loss
        loss = None
        if targets is not None:
            lidar_targets, state_targets = targets
            rec_loss = F.mse_loss(lidar_rec, lidar)  # input-reconstruction loss
            pred_loss = F.mse_loss(lidar_pred, lidar_targets)  # reconstructed prediction loss
            state_loss = F.mse_loss(state_pred, state_targets) / STATE_NORM_FACTOR**2
            KLD = -0.5 * torch.mean(1 + logvar - mu.pow(2) - logvar.exp())  # kullback leibler
            kld_tolerance = 0.5
            kld_weight = 0.001
            KLD = torch.max(KLD, kld_tolerance * torch.ones_like(KLD))
            loss = rec_loss + kld_weight * KLD + pred_loss + state_loss

        return lidar_pred, state_pred, loss

    def _to_correct_device(self, tensor):
        if self.gpu:
            if torch.cuda.is_available():
                device = torch.cuda.current_device()
                return tensor.to(device)
            else:
                print("WARNING: model created with gpu enabled, but no gpu found")
        return tensor

    def encode(self, lidar):
        """
        lidar: numpy (batch, L, CH)

        OUTPUTS
        z: (batch, Z)
        """
        b, L, CH = lidar.shape

        lidar_t = torch.tensor(np.moveaxis(lidar, -1, 1), dtype=torch.float)
        lidar_t = self._to_correct_device(lidar_t)

        z, mu, logvar = self.convVAE.encode(lidar_t)
        return z.detach().cpu().numpy()

    def encode_mu_logvar(self, lidar):
        """
        lidar: numpy (batch, L, CH)


        OUTPUTS
        mu: (batch, Z)
        logvar: (batch, Z)
        """
        b, L, CH = lidar.shape

        lidar_t = torch.tensor(np.moveaxis(lidar, -1, 1), dtype=torch.float)
        lidar_t = self._to_correct_device(lidar_t)

        z, mu, logvar = self.convVAE.encode(lidar_t)
        mu = mu.detach().cpu().numpy()
        logvar = logvar.detach().cpu().numpy()
        return mu, logvar

    def decode(self, z):
        """
        z: numpy (batch, Z)

        OUTPUTS
        lidar_rec: (batch, L, CH)
        """
        b, Z = z.shape

        z_t = torch.tensor(z, dtype=torch.float)
        z_t = self._to_correct_device(z_t)

        lidar_rec_t = self.convVAE.decode(z_t) # b, CH, L
        lidar_rec = np.moveaxis(lidar_rec_t.detach().cpu().numpy(), 1, -1)
        return lidar_rec

    def get_h(self, gpt_sequence):
        """ for compat with encodedenv
        gpt sequence is a list of dicts, one for each step in the sequence.
        each dict has
        "obs": numpy (L, CH) [0, 1]
        "state": numpy (2,) [-inf, inf]
        "action": numpy (3,) [-inf, inf]
        """
        _b = 1  # batch size
        lidar = np.array([d["obs"] for d in gpt_sequence])  # t, L, CH
        lidar = np.moveaxis(lidar, -1, 1)
        lidar = lidar.reshape((_b, *lidar.shape))
        lidar_t = torch.tensor(lidar, dtype=torch.float)
        lidar_t = self._to_correct_device(lidar_t)
        state = np.array([d["state"] for d in gpt_sequence])  # t, 2
        state = state.reshape((_b, *state.shape))
        state_t = torch.tensor(state, dtype=torch.float)
        state_t = self._to_correct_device(state_t)
        action = np.array([d["action"] for d in gpt_sequence])  # t, 3
        action = action.reshape((_b, *action.shape))
        action_t = torch.tensor(action, dtype=torch.float)
        action_t = self._to_correct_device(action_t)
        dones = np.zeros((_b, len(gpt_sequence), 1))
        dones_t = torch.tensor(dones, dtype=torch.float)
        dones_t = self._to_correct_device(dones_t)
        h_container = [None]
        self.forward(lidar_t, state_t, action_t, dones_t, h=h_container)
        h = h_container[0].detach().cpu().numpy()
        h = h[0, -1]  # only batch, last item in sequence
        return h



import os
from pyniel.python_tools.path_tools import make_dir_if_not_exists

from navrep.tools.commonargs import parse_common_args
from navrep.envs.e2eenv import E2EIANEnv
from navrep.scripts.cross_test_navreptrain_in_ianenv import run_test_episodes
from navrep.scripts.test_e2e import E2ECPolicy


if __name__ == '__main__':
    args, _ = parse_common_args()

    if args.n is None:
        args.n = 1000
    collect_trajectories = False

    env = E2EIANEnv(silent=True, collect_trajectories=collect_trajectories)
    policy = E2ECPolicy()

    S = run_test_episodes(env, policy, render=args.render, num_episodes=args.n)

    DIR = os.path.expanduser("~/navrep/eval/crosstest")
    if args.dry_run:
        DIR = "/tmp/navrep/eval/crosstest"
    make_dir_if_not_exists(DIR)
    if collect_trajectories:
        NAME = "e2enavreptrain_in_ianenv_{}.pckl".format(len(S))
        PATH = os.path.join(DIR, NAME)
        S.to_pickle(PATH)
    else:
        NAME = "e2enavreptrain_in_ianenv_{}.csv".format(len(S))
        PATH = os.path.join(DIR, NAME)
        S.to_csv(PATH)
    print("{} written.".format(PATH))



import numpy as np
from matplotlib import pyplot as plt

O_TRAJ_PROB = 0.01  # 1 means all trajectories are shown. lowering this number randomly removes trajs
S = 30  # markersize in points


def truncate_colormap(cmap, minval=0.0, maxval=1.0, n=100):
    # from stackoverflow how-to-extract-a-subset-of-a-colormap-as-a-new-colormap-in-matplotlib
    from matplotlib import colors
    new_cmap = colors.LinearSegmentedColormap.from_list(
        'trunc({n},{a:.2f},{b:.2f})'.format(n=cmap.name, a=minval, b=maxval),
        cmap(np.linspace(minval, maxval, n)))
    return new_cmap

def blue(u=np.random.rand()):
    cmap = truncate_colormap(plt.cm.bwr, 0., .4)
    u = np.clip(u, 0., 1.)
    return cmap(u)

def orange(u=np.random.rand()):
    cmap = truncate_colormap(plt.cm.Wistia, .3, .7)
    u = np.clip(u, 0., 1.)
    return cmap(u)

def plot_ianenv_statistics(S, ianenv, focus_scenario=None):
    plt.ion()
    if focus_scenario is None:
        scenarios = sorted(list(set(S["scenario"].values)))
    else:
        scenarios = [focus_scenario]
    if len(scenarios) > 6:
        fig, axes = plt.subplots(5, 6)
    else:
        fig, axes = plt.subplots(len(scenarios), 1)
    axes = np.array(axes).reshape((-1))
    for i, scenario in enumerate(scenarios):
        is_scenario = S["scenario"].values == scenario
        scenario_S = S[is_scenario]
        # print results
        goal_reached = scenario_S["goal_reached"].values
        goal_reached_perc = np.sum(goal_reached) * 100. / len(goal_reached)
        timed_out = scenario_S["timed_out"].values
        timed_out_perc = np.sum(timed_out) * 100. / len(timed_out)
        collisioned_out = scenario_S["collisioned_out"].values
        collisioned_out_perc = np.sum(collisioned_out) * 100. / len(collisioned_out)
        print("{} -----------------------------------".format(scenario))
        print("Goal reached: {}%".format(goal_reached_perc))
        print("Timed out: {}%".format(timed_out_perc))
        print("Collisioned out: {}%".format(collisioned_out_perc))

        # plot result
        trajectories = scenario_S["trajectory"]
        o_trajectories = scenario_S["other_trajectories"]
        goals = scenario_S["goal"]

        ax = axes[i]

        # TODO make this a map2d utility function
        # TODO get scenario map
        ianenv.reset(set_scenario=scenario)
        contours = ianenv.iarlenv.rlenv.virtual_peppers[0].map2d.as_closed_obst_vertices()
        for c in contours:
            cplus = np.concatenate((c, c[:1, :]), axis=0)
            ax.plot(cplus[:,0], cplus[:,1], color='k')
        # ^^^^

        for t, g, s, ot in zip(trajectories, goals, goal_reached, o_trajectories):
            line_color = blue(len(t)/1000.) if s else orange(len(t)/1000.)
            zorder = 2 if s else 1
            ax.plot(t[:,0], t[:,1], color=line_color, zorder=zorder)
            for o in ot:
                if len(o) > 0:
                    if np.random.rand() < O_TRAJ_PROB:
                        for xy in o:
                            ax.add_artist(plt.Circle((xy), 0.3,
                                                     edgecolor="dimgray", facecolor="none",
                                                     alpha=0.1, zorder=0))
            ax.add_artist(plt.Circle((t[0,0], t[0,1]), 0.3, color="green", zorder=2))
            ax.add_artist(plt.Circle((g[0], g[1]), 0.3, color="red", zorder=2))
        if focus_scenario is None:
            ax.set_title(scenario)
        else:
            ax.set_title("{} - {} - {}%".format(scenario, log, goal_reached_perc))
        ax.axis("equal")
        ax.set_adjustable('box')
        if "rl" in scenario:
            xlim = (np.min([np.min(t[:,0]) for t in trajectories]) - 3,
                    np.max([np.max(t[:,0]) for t in trajectories]) + 3)
            ylim = (np.min([np.min(t[:,1]) for t in trajectories]) - 3,
                    np.max([np.max(t[:,1]) for t in trajectories]) + 3)
            ax.set(xlim=xlim, ylim=ylim)

    goal_reached = S["goal_reached"].values
    goal_reached_perc = np.sum(goal_reached) * 100. / len(goal_reached)
    timed_out = S["timed_out"].values
    timed_out_perc = np.sum(timed_out) * 100. / len(timed_out)
    collisioned_out = S["collisioned_out"].values
    collisioned_out_perc = np.sum(collisioned_out) * 100. / len(collisioned_out)
    print("All scenarios -----------------------------------")
    print("Goal reached: {}%".format(goal_reached_perc))
    print("Timed out: {}%".format(timed_out_perc))
    print("Collisioned out: {}%".format(collisioned_out_perc))
    plt.pause(1.)


if __name__ == "__main__":
    import os
    import pandas as pd

    from navrep.tools.commonargs import parse_plotting_args
    from navrep.envs.ianenv import IANEnv

    args, _ = parse_plotting_args()

    LOGDIR = args.logdir
    if LOGDIR is None:
        LOGDIR = "~/navrep"
    LOGDIR = os.path.join(LOGDIR, "eval/crosstest")
    LOGDIR = os.path.expanduser(LOGDIR)
    logs = sorted(os.listdir(LOGDIR))
    for log in logs:
        if not log.endswith(".pckl"):
            continue
        print()
        print(log)
        path = os.path.join(LOGDIR, log)
        S = pd.read_pickle(path)

        env = IANEnv(silent=True)

        plot_ianenv_statistics(S, env, focus_scenario=args.scenario)



from __future__ import print_function
import numpy as np
import os
from pyniel.python_tools.path_tools import make_dir_if_not_exists

from navrep.models.vae2d import ConvVAE, reset_graph
from navrep.tools.rings import generate_rings
from navrep.tools.commonargs import parse_common_args

common_args, _ = parse_common_args()
VARIANT = common_args.environment
print(common_args)
_Z = 32

if VARIANT == "ian":
    SS = 5  # sub-sample dataset to increase DT
    vae_path = os.path.expanduser("~/navrep/models/V/vae.json")
    V_dataset_folder = os.path.expanduser("~/navrep/datasets/V/ian")
    M_dataset_folder = os.path.expanduser("~/navrep/datasets/M/ian")
if VARIANT == "toy":
    SS = None  # no need to sub-sample, toy env is already at DT=0.5
    vae_path = os.path.expanduser("~/navrep/models/V/toyvae.json")
    V_dataset_folder = os.path.expanduser("~/navrep/datasets/V/toy")
    M_dataset_folder = os.path.expanduser("~/navrep/datasets/M/toy")
if VARIANT == "markone":
    SS = None
    vae_path = os.path.expanduser("~/navrep/models/V/markonevae.json")
    V_dataset_folder = os.path.expanduser("~/navrep/datasets/V/markone")
    M_dataset_folder = os.path.expanduser("~/navrep/datasets/M/markone")
if VARIANT == "marktwo":
    SS = None
    vae_path = os.path.expanduser("~/navrep/models/V/marktwovae.json")
    V_dataset_folder = os.path.expanduser("~/navrep/datasets/V/marktwo")
    M_dataset_folder = os.path.expanduser("~/navrep/datasets/M/marktwo")
if VARIANT == "navreptrain":
    SS = None
    vae_path = os.path.expanduser("~/navrep/models/V/navreptrainvae.json")
    V_dataset_folder = os.path.expanduser("~/navrep/datasets/V/navreptrain")
    M_dataset_folder = os.path.expanduser("~/navrep/datasets/M/navreptrain")


# create network
reset_graph()
vae = ConvVAE(z_size=_Z, batch_size=1, is_training=False,)
vae.load_json(vae_path)

# rings converter
rings_def = generate_rings(64, 64)

# labels to learn are x, r, d (obs, reward, done)

files = []
for dirpath, dirnames, filenames in os.walk(V_dataset_folder):
    for filename in [f for f in filenames if f.endswith(".npz")]:
        files.append(os.path.join(dirpath, filename))
files = sorted(files)
for path in files:
    arrays = np.load(path)
    scans = arrays["scans"]
    robotstates = arrays["robotstates"]
    rewards = arrays["rewards"]
    actions = arrays["actions"]
    dones = arrays["dones"]

    if SS is not None:
        scans = np.concatenate([scans[i::SS] for i in range(SS)], axis=0)
        robotstates = np.concatenate([robotstates[i::SS] for i in range(SS)], axis=0)
        rewards = np.concatenate([rewards[i::SS] for i in range(SS)], axis=0)
        actions = np.concatenate([actions[i::SS] for i in range(SS)], axis=0)
        dones_ss = [dones[i::SS] for i in range(SS)]
        for d in dones_ss:
            d[-1] = True
        dones = np.concatenate(dones_ss, axis=0)

    obs = rings_def["lidar_to_rings"](scans).astype(float) / rings_def["rings_to_bool"]

    mus = []
    logvars = []
    batch_size = 100
    for i in range(0, len(obs), batch_size):
        mu, logvar = vae.encode_mu_logvar(obs[i : i + batch_size, :, :, :])
        z = mu + np.exp(logvar / 2.0) * np.random.randn(*logvar.shape)
        mus.extend(mu)
        logvars.extend(logvar)
    mus = np.array(mus).astype(np.float32)
    logvars = np.array(logvars).astype(np.float32)
    savepath = path.replace(V_dataset_folder, M_dataset_folder).replace("scans", "mus_logvars")
    make_dir_if_not_exists(os.path.dirname(savepath))
    np.savez_compressed(
        savepath,
        mus=mus,
        logvars=logvars,
        robotstates=robotstates,
        rewards=rewards,
        actions=actions,
        dones=dones,
    )
    print("z saved to {}".format(savepath))



from __future__ import print_function
import numpy as np
import os
from datetime import datetime
import random
import time
import pickle
from pyniel.python_tools.path_tools import make_dir_if_not_exists
import pandas as pd

from navrep.models.vae2d import ConvVAE
from navrep.models.vae1d import Conv1DVAE
from navrep.models.rnn import reset_graph, default_hps, MDNRNN, MAX_GOAL_DIST
from navrep.tools.test_worldmodel import rnn_worldmodel_error, vae1d_rnn_worldmodel_error
from navrep.tools.commonargs import parse_common_args
from navrep.scripts.train_vae import _Z

_H = 512
_G = 2  # goal states
_A = 3  # action dims

if __name__ == "__main__":
    common_args, _ = parse_common_args()
    VARIANT = common_args.environment

    START_TIME = datetime.now().strftime("%Y_%m_%d__%H_%M_%S")
    MAX_STEPS = common_args.n
    if MAX_STEPS is None:
        MAX_STEPS = 222222
    N_EPOCHS = MAX_STEPS  # don't limit based on epoch
    VAE_TYPE = "1d" if common_args.backend == "VAE1D_LSTM" else ""

    hps = default_hps()
    hps = hps._replace(seq_width=_Z+_G, action_width=_A, rnn_size=_H)
    print(hps)
    # hps.batch_size = 100
    # hps.max_seq_len = 1000
    # hps = hps._replace(learning_rate=0.0001)

    if VARIANT == "ian":
        dataset_folder = os.path.expanduser("~/navrep/datasets/M/ian")
        test_dataset_folder = os.path.expanduser("~/navrep/datasets/V/ian")
        log_path = os.path.expanduser("~/navrep/logs/M/rnn_train_log_{}.csv".format(START_TIME))
        log_hyperparams_path = os.path.expanduser(
            "~/navrep/logs/M/rnn_train_log_{}.hyperparams.pckl".format(START_TIME))
        model_hyperparams_path = os.path.expanduser("~/navrep/models/M/rnn.hyperparams.pckl")
        model_path = os.path.expanduser("~/navrep/models/M/rnn.json")
        vae_model_path = os.path.expanduser("~/navrep/models/V/vae.json")
    if VARIANT == "toy":
        dataset_folder = os.path.expanduser("~/navrep/datasets/M/toy")
        test_dataset_folder = os.path.expanduser("~/navrep/datasets/V/toy")
        log_path = os.path.expanduser("~/navrep/logs/M/toyrnn_train_log_{}.csv".format(START_TIME))
        log_hyperparams_path = os.path.expanduser(
            "~/navrep/logs/M/toyrnn_train_log_{}.hyperparams.pckl".format(START_TIME))
        model_hyperparams_path = os.path.expanduser("~/navrep/models/M/toyrnn.hyperparams.pckl")
        model_path = os.path.expanduser("~/navrep/models/M/toyrnn.json")
        vae_model_path = os.path.expanduser("~/navrep/models/V/toyvae.json")
    if VARIANT == "markone":
        dataset_folder = os.path.expanduser("~/navrep/datasets/M/markone")
        test_dataset_folder = os.path.expanduser("~/navrep/datasets/V/markone")
        log_path = os.path.expanduser("~/navrep/logs/M/markonernn_train_log_{}.csv".format(START_TIME))
        log_hyperparams_path = os.path.expanduser(
            "~/navrep/logs/M/markonernn_train_log_{}.hyperparams.pckl".format(START_TIME))
        model_hyperparams_path = os.path.expanduser("~/navrep/models/M/markonernn.hyperparams.pckl")
        model_path = os.path.expanduser("~/navrep/models/M/markonernn.json")
        vae_model_path = os.path.expanduser("~/navrep/models/V/markonevae.json")
    if VARIANT == "marktwo":
        dataset_folder = os.path.expanduser("~/navrep/datasets/M/marktwo")
        test_dataset_folder = os.path.expanduser("~/navrep/datasets/V/marktwo")
        log_path = os.path.expanduser("~/navrep/logs/M/marktwornn_train_log_{}.csv".format(START_TIME))
        log_hyperparams_path = os.path.expanduser(
            "~/navrep/logs/M/marktwornn_train_log_{}.hyperparams.pckl".format(START_TIME))
        model_hyperparams_path = os.path.expanduser("~/navrep/models/M/marktwornn.hyperparams.pckl")
        model_path = os.path.expanduser("~/navrep/models/M/marktwornn.json")
        vae_model_path = os.path.expanduser("~/navrep/models/V/marktwovae.json")
    if VARIANT == "navreptrain":
        dataset_folder = os.path.expanduser("~/navrep/datasets/M/navreptrain")
        test_dataset_folder = os.path.expanduser("~/navrep/datasets/V/navreptrain")
        log_path = os.path.expanduser(
            "~/navrep/logs/M/navreptrainrnn{}_train_log_{}.csv".format(VAE_TYPE, START_TIME))
        log_hyperparams_path = os.path.expanduser(
            "~/navrep/logs/M/navreptrainrnn{}_train_log_{}.hyperparams.pckl".format(VAE_TYPE, START_TIME))
        model_hyperparams_path = os.path.expanduser(
            "~/navrep/models/M/navreptrainrnn{}.hyperparams.pckl".format(VAE_TYPE))
        model_path = os.path.expanduser("~/navrep/models/M/navreptrainrnn{}.json".format(VAE_TYPE))
        vae_model_path = os.path.expanduser("~/navrep/models/V/navreptrainvae{}.json".format(VAE_TYPE))

    if common_args.dry_run:
        log_path = log_path.replace(os.path.expanduser("~/navrep"), "/tmp/navrep")
        log_hyperparams_path = log_hyperparams_path.replace(os.path.expanduser("~/navrep"), "/tmp/navrep")
        model_path = model_path.replace(os.path.expanduser("~/navrep"), "/tmp/navrep")
        model_hyperparams_path = model_hyperparams_path.replace(os.path.expanduser("~/navrep"), "/tmp/navrep")

    make_dir_if_not_exists(os.path.dirname(model_path))
    make_dir_if_not_exists(os.path.dirname(log_path))

    # load preprocessed data
    files = []
    for dirpath, dirnames, filenames in os.walk(dataset_folder):
        for filename in [f for f in filenames if f.endswith(".npz")]:
            files.append(os.path.join(dirpath, filename))
    all_data = []
    for path in files:
        arrays = np.load(path)
        all_data.append(
            [
                arrays["mus"],
                arrays["logvars"],
                arrays["robotstates"],
                arrays["actions"],
                arrays["dones"],
                arrays["rewards"],
            ]
        )
    n_total_frames = np.sum([mu.shape[0] for mu, _, _, _, _, _ in all_data])
    chunksize = hps.batch_size * hps.max_seq_len  # frames per batch (100'000)
    print("total frames: ", n_total_frames)
    if n_total_frames < chunksize:
        raise ValueError()

    reset_graph()
    model = MDNRNN(hps)
    model.print_trainable_params()
    vae = None

    viewer = None
    values_logs = None

    start = time.time()
    for epoch in range(1, N_EPOCHS + 1):
        #     print('preparing data for epoch', epoch)
        batches_start = time.time()
        # flatten all sequences into one
        mu_sequence = np.zeros((n_total_frames, _Z), dtype=np.float32)
        logvar_sequence = np.zeros((n_total_frames, _Z), dtype=np.float32)
        robotstate_sequence = np.zeros((n_total_frames, 5), dtype=np.float32)
        action_sequence = np.zeros((n_total_frames, 3), dtype=np.float32)
        done_sequence = np.zeros((n_total_frames, 1), dtype=np.float32)
        reward_sequence = np.zeros((n_total_frames, 1), dtype=np.float32)
        i = 0
        random.shuffle(all_data)
        for mu, logvar, robotstate, action, done, reward in all_data:
            L = len(mu)
            mu_sequence[i : i + L, :] = mu.reshape(L, _Z)
            logvar_sequence[i : i + L, :] = logvar.reshape(L, _Z)
            robotstate_sequence[i : i + L, :] = robotstate.reshape(L, 5)
            action_sequence[i : i + L, :] = action.reshape(L, 3)
            done_sequence[i : i + L, :] = done.reshape(L, 1)
            reward_sequence[i : i + L, :] = reward.reshape(L, 1)
            i += L
        # sample z from  mu and logvar
        z_sequence = mu_sequence + np.exp(logvar_sequence / 2.0) * np.random.randn(
            *(mu_sequence.shape)
        )
        # add goalstate (robotstate[:2]) to z
        robotstate_sequence[:, :_G] = robotstate_sequence[:, :_G] / MAX_GOAL_DIST  # normalize goal dist
        z_rs_sequence = np.concatenate([z_sequence, robotstate_sequence[:, :_G]], axis=-1)
        # resize array to be reshapable into sequences and batches
        n_chunks = n_total_frames // chunksize
        # reshape into sequences
        z_rs_sequences = np.reshape(
            z_rs_sequence[: n_chunks * chunksize, :], (-1, hps.max_seq_len, _Z+_G)
        )
        action_sequences = np.reshape(
            action_sequence[: n_chunks * chunksize], (-1, hps.max_seq_len, 3)
        )
        done_sequences = np.reshape(
            done_sequence[: n_chunks * chunksize], (-1, hps.max_seq_len)
        )
        reward_sequences = np.reshape(
            reward_sequence[: n_chunks * chunksize], (-1, hps.max_seq_len)
        )
        num_sequences = len(z_rs_sequences)
        if num_sequences == 0:
            raise ValueError("Not enough data for a single batch")
        # shuffle
        random_idxs = list(range(num_sequences))
        random.shuffle(random_idxs)
        random_idxs = np.reshape(random_idxs, (-1, hps.batch_size))
        # reshape into batches
        z_rs_batches = z_rs_sequences[random_idxs]
        action_batches = action_sequences[random_idxs]
        done_batches = done_sequences[random_idxs]
        reward_batches = reward_sequences[random_idxs]
        num_batches = len(z_rs_batches)
        # result is of size (n_batches, batch_size, seq_len, ...)
        #     print('number of batches', num_batches)
        batches_end = time.time()
        batch_time_taken = batches_end - batches_start
        #     print('time taken to create batches', batch_time_taken)

        batch_state = model.sess.run(model.initial_state)

        for batch_z_rs, batch_action, batch_done, batch_reward in zip(
            z_rs_batches, action_batches, done_batches, reward_batches
        ):
            step = model.sess.run(model.global_step)
            curr_learning_rate = (hps.learning_rate - hps.min_learning_rate) * (
                hps.decay_rate
            ) ** step + hps.min_learning_rate

            feed = {
                model.batch_z_rs: batch_z_rs,
                model.batch_action: batch_action,
                model.batch_restart: batch_done,
                model.initial_state: batch_state,
                model.lr: curr_learning_rate,
            }

            (train_cost, z_cost, r_cost, batch_state, train_step, _) = model.sess.run(
                [
                    model.cost,
                    model.z_cost,
                    model.r_cost,
                    model.final_state,
                    model.global_step,
                    model.train_op,
                ],
                feed,
            )

            lidar_e = None
            state_e = None
            if step % 200 == 0:
                # load VAE
                if VAE_TYPE == "1d":
                    if vae is None:
                        vae = Conv1DVAE(z_size=_Z, batch_size=model.hps.max_seq_len-1, is_training=False)
                        vae.load_json(vae_model_path)
                    lidar_e, state_e = vae1d_rnn_worldmodel_error(model, test_dataset_folder, vae)
                else:
                    if vae is None:
                        vae = ConvVAE(z_size=_Z, batch_size=model.hps.max_seq_len-1, is_training=False)
                        vae.load_json(vae_model_path)
                    lidar_e, state_e = rnn_worldmodel_error(model, test_dataset_folder, vae)

                print("Test: lidar error {}, state error {}".format(lidar_e, state_e))
                model.save_json(model_path)

            if step % 20 == 0 and step > 0:
                end = time.time()
                time_taken = end - start
                start = time.time()
                output_log = (
                    "step: %d, lr: %.6f, cost: %.4f, z_cost: %.4f, r_cost: %.4f, train_time_taken: %.4f"
                    % (step, curr_learning_rate, train_cost, z_cost, r_cost, time_taken)
                )
                print(output_log)
                # log
                values_log = pd.DataFrame(
                    [[step, curr_learning_rate, train_cost, z_cost, r_cost, time_taken, lidar_e, state_e]],
                    columns=["step", "lr", "cost", "z_cost", "r_cost", "train_time_taken",
                             "lidar_test_error", "state_test_error"],
                )
                if values_logs is None:
                    values_logs = values_log.copy()
                else:
                    values_logs = values_logs.append(values_log, ignore_index=True)
                values_logs.to_csv(log_path)
                with open(log_hyperparams_path, "wb") as f:
                    pickle.dump(hps, f)

            if common_args.render:  # Visually check that the batch is sound
                import matplotlib.pyplot as plt
                from navrep.tools.rings import generate_rings

                reset_graph()
                vae = ConvVAE(z_size=_Z, batch_size=1, is_training=False)
                vae.load_json(vae_model_path)
                rings_def = generate_rings(64, 64)
                rings_pred = vae.decode(batch_z_rs[0, :, :_Z]) * rings_def["rings_to_bool"]
                plt.ion()
                for i, ring in enumerate(rings_pred):
                    rings_def["visualize_rings"](ring, scan=None)
                    plt.scatter(batch_z_rs[0, i, _Z], batch_z_rs[0, i, 33], color='red')
                    plt.ylim([0, 10])
                    plt.title("{:.1f} {:.1f} {:.1f}".format(*batch_action[0, i]))
                    plt.pause(0.5)
                exit()
            if False:  # render all sequences in batch at once
                from navrep.tools.render import render_lidar_batch
                from navrep.tools.rings import generate_rings

                reset_graph()
                vae = ConvVAE(z_size=_Z, batch_size=100, is_training=False)
                vae.load_json(vae_model_path)
                rings_def = generate_rings(64, 64)
                batch_decodings = []
                for i in range(batch_z_rs.shape[1]):  # for each sequence step
                    rings_pred = vae.decode(batch_z_rs[:, i, :_Z]) * rings_def["rings_to_bool"]
                    predicted_ranges = rings_def["rings_to_lidar"](rings_pred, 1080)
                    batch_decodings.append(predicted_ranges)
                for i, predicted_ranges in enumerate(batch_decodings):
                    viewer = render_lidar_batch(
                        predicted_ranges, 0, 2 * np.pi, viewer=viewer
                    )
                    import pyglet

                    filename = "/tmp/frame{:03}.png".format(i)
                    pyglet.image.get_buffer_manager().get_color_buffer().save(filename)
                    print("image file writen : ", filename)

            if MAX_STEPS is not None:
                if train_step > MAX_STEPS:
                    break
        if MAX_STEPS is not None:
            if train_step > MAX_STEPS:
                break

    model.save_json(model_path)



import os
import matplotlib
if os.path.expandvars("$MACHINE_NAME") in ["leonhard", "euler"]:
    matplotlib.use('agg')
import logging
import os
import math
import time
from datetime import datetime
import numpy as np
import pandas as pd
from tqdm import tqdm
import torch
import torch.optim as optim
from torch.utils.data.dataloader import DataLoader
from pyniel.python_tools.path_tools import make_dir_if_not_exists

from navrep.models.vae1dlstm import VAE1DLSTM, VAE1DLSTMConfig
from navrep.models.gpt import save_checkpoint, set_seed
from navrep.tools.wdataset import WorldModelDataset
from navrep.tools.test_worldmodel import vae1dlstm_worldmodel_error
from navrep.tools.commonargs import parse_common_args

_Z = 32
_H = 64
_S = 64  # sequence length


if __name__ == "__main__":
    args, _ = parse_common_args()

    START_TIME = datetime.now().strftime("%Y_%m_%d__%H_%M_%S")

    if args.environment == "navreptrain":
        dataset_dir = os.path.expanduser("~/navrep/datasets/V/navreptrain")
        data_regen = "navreptrain"
        log_path = os.path.expanduser("~/navrep/logs/W/navreptrain_vae1dlstm_train_log_{}.csv".format(START_TIME))
        checkpoint_path = os.path.expanduser("~/navrep/models/W/navreptrainvae1dlstm")
    else:
        raise NotImplementedError(args.environment)

    if args.dry_run:
        log_path = log_path.replace(os.path.expanduser("~/navrep"), "/tmp/navrep")
        checkpoint_path = checkpoint_path.replace(os.path.expanduser("~/navrep"), "/tmp/navrep")

    make_dir_if_not_exists(os.path.dirname(checkpoint_path))
    make_dir_if_not_exists(os.path.dirname(log_path))
    make_dir_if_not_exists(os.path.expanduser("~/tmp_navrep"))

    # make deterministic
    set_seed(42)

    # set up logging
    logging.basicConfig(
        format="%(asctime)s - %(levelname)s - %(name)s -   %(message)s",
        datefmt="%m/%d/%Y %H:%M:%S",
        level=logging.INFO,
    )
    logger = logging.getLogger(__name__)

    mconf = VAE1DLSTMConfig(_Z, _H)

    train_dataset = WorldModelDataset(
        dataset_dir, _S,
        lidar_mode="scans",
        pre_convert_obs=True,
        regen=data_regen,
    )

    # training params
    # optimization parameters
    max_steps = args.n
    if max_steps is None:
        max_steps = 222222
    max_epochs = max_steps  # don't stop based on epoch
    batch_size = 128
    learning_rate = 6e-4
    betas = (0.9, 0.95)
    grad_norm_clip = 1.0
    lr_decay = True  # learning rate decay params: linear warmup followed by cosine decay to 10% of original
    weight_decay = 0.1  # only applied on matmul weights
    warmup_tokens = 512 * 20
    final_tokens = 200 * len(train_dataset) * _S
    num_workers = 0  # for DataLoader

    # create model
    model = VAE1DLSTM(mconf)

    # take over whatever gpus are on the system
    device = "cpu"
    if torch.cuda.is_available():
        device = torch.cuda.current_device()
        model = torch.nn.DataParallel(model).to(device)

    # create the optimizer
    no_decay = ["bias", "LayerNorm.weight"]
    params_decay = [
        p for n, p in model.named_parameters() if not any(nd in n for nd in no_decay)
    ]
    params_nodecay = [
        p for n, p in model.named_parameters() if any(nd in n for nd in no_decay)
    ]
    optim_groups = [
        {"params": params_decay, "weight_decay": weight_decay},
        {"params": params_nodecay, "weight_decay": 0.0},
    ]
    optimizer = optim.AdamW(optim_groups, lr=learning_rate, betas=betas)

    global_step = 0
    tokens = 0  # counter used for learning rate decay
    values_logs = None
    start = time.time()
    for epoch in range(max_epochs):
        is_train = True
        model.train(is_train)
        loader = DataLoader(
            train_dataset,
            shuffle=is_train,
            batch_size=batch_size,
            num_workers=num_workers,
        )

        losses = []
        pbar = tqdm(enumerate(loader), total=len(loader)) if is_train else enumerate(loader)
        for it, (x, a, y, x_rs, y_rs, dones) in pbar:
            global_step += 1

            # initialize sequence rnn_state
            this_batch_size = x.shape[0]  # last batch can be smaller
            h0 = torch.randn(mconf.lstm.n_layer, this_batch_size, mconf.lstm.h_size)
            c0 = torch.randn(mconf.lstm.n_layer, this_batch_size, mconf.lstm.h_size)

            # place data on the correct device
            x = x.to(device)
            x_rs = x_rs.to(device)
            a = a.to(device)
            y = y.to(device)
            y_rs = y_rs.to(device)
            dones = dones.to(device)
            h0 = h0.to(device)
            c0 = c0.to(device)

            # forward the model
            with torch.set_grad_enabled(is_train):
                y_pred, y_rs_pred, loss = model(x, x_rs, a, dones, [h0, c0], targets=(y, y_rs))
                loss = (
                    loss.mean()
                )  # collapse all losses if they are scattered on multiple gpus
                losses.append(loss.item())

            if is_train:

                # backprop and update the parameters
                model.zero_grad()
                loss.backward()
                torch.nn.utils.clip_grad_norm_(model.parameters(), grad_norm_clip)
                optimizer.step()

                # decay the learning rate based on our progress
                if lr_decay:
                    tokens += (
                        a.shape[0] * a.shape[1]
                    )  # number of tokens processed this step
                    if tokens < warmup_tokens:
                        # linear warmup
                        lr_mult = float(tokens) / float(max(1, warmup_tokens))
                    else:
                        # cosine learning rate decay
                        progress = float(tokens - warmup_tokens) / float(
                            max(1, final_tokens - warmup_tokens)
                        )
                        lr_mult = max(0.1, 0.5 * (1.0 + math.cos(math.pi * progress)))
                    lr = learning_rate * lr_mult
                    for param_group in optimizer.param_groups:
                        param_group["lr"] = lr
                else:
                    lr = learning_rate

                # report progress
                pbar.set_description(
                    f"epoch {epoch}: train loss {loss.item():.5f}. lr {lr:e}"
                )

                if global_step == 1 or global_step % 1000 == 0:
                    # save plot
                    from matplotlib import pyplot as plt
                    plt.figure("training_status")
                    plt.clf()
                    plt.suptitle("training step {}".format(global_step))
                    f, axes = plt.subplots(3, 5, num="training_status", sharex=True, sharey=True)
                    for i, (ax0, ax1, ax2) in enumerate(axes.T):
                        ax0.plot(np.squeeze(x.cpu()[0, 5 + i]))
                        ax1.plot(np.squeeze(y.cpu()[0, 5 + i]))
                        ax2.plot(np.squeeze(y_pred.detach().cpu()[0, 5 + i]))
                        ax2.set_xlabel("Done {}".format(dones.cpu()[0, 5 + 1]))
                    plt.savefig(os.path.expanduser(
                        "~/tmp_navrep/vae1dlstm_step{:07}.png").format(global_step))

        lidar_e = None
        state_e = None
        if epoch % 20 == 0:
            lidar_e, state_e = vae1dlstm_worldmodel_error(model, dataset_dir, device)
            save_checkpoint(model, checkpoint_path)

        # log
        end = time.time()
        time_taken = end - start
        start = time.time()
        values_log = pd.DataFrame(
            [[global_step, loss.item(), lidar_e, state_e, time_taken]],
            columns=["step", "cost", "lidar_test_error", "state_test_error", "train_time_taken"],
        )
        if values_logs is None:
            values_logs = values_log.copy()
        else:
            values_logs = values_logs.append(values_log, ignore_index=True)
        if log_path is not None:
            values_logs.to_csv(log_path)

        if not is_train:
            logger.info("test loss: %f", np.mean(losses))

        if global_step >= max_steps:
            break



import os

from navrep.envs.markencodedenv import MarkTwoEncodedEnv
from navrep.envs.markenv import SECOND_TRAIN_MAPS, SECOND_TEST_MAPS
from navrep.scripts.test_mark_common import MarkTwoCPolicy, markeval

if __name__ == "__main__":
    os.environ["CUDA_VISIBLE_DEVICES"] = "-1"  # disable GPU

    policy = MarkTwoCPolicy()

    env = MarkTwoEncodedEnv(silent=True, maps=SECOND_TRAIN_MAPS)
    S1 = markeval(policy, env, n_episodes=100, render=False)

    env = MarkTwoEncodedEnv(silent=True, maps=SECOND_TEST_MAPS)
    S2 = markeval(policy, env, n_episodes=100, render=False)

    import matplotlib.pyplot as plt
    plt.ioff()
    plt.show()



import os
import numpy as np
from tqdm import tqdm
from stable_baselines import PPO2
from navrep.envs.navreptrainencodedenv import NavRepTrainEncodedEnv

from crowd_sim.envs.utils.info import Timeout, ReachGoal, Danger, Collision, CollisionOtherAgent
from navrep.tools.commonargs import parse_common_args

class NavRepCPolicy(object):
    """ wrapper for gym policies """
    def __init__(self, model=None):
        if model is not None:
            self.model = model
        else:
            self.model_path = os.path.expanduser(
                "~/navrep/models/gym/navreptrainencodedenv_latest_PPO_ckpt")
            self.model = PPO2.load(self.model_path)
            print("Model '{}' loaded".format(self.model_path))

    def act(self, obs):
        action, _states = self.model.predict(obs, deterministic=True)
        return action


def run_test_episodes(env, policy, render=False, print_failure=True, num_episodes=500):
    success_times = []
    collision_times = []
    collision_other_agent_times = []
    timeout_times = []
    success = 0
    collision = 0
    collision_other_agent = 0
    timeout = 0
    too_close = 0
    min_dist = []
    cumulative_rewards = []
    collision_cases = []
    collision_other_agent_cases = []
    timeout_cases = []
    progress_bar = tqdm(range(num_episodes), total=num_episodes)
    for i in progress_bar:
        progress_bar.set_description("Case {}".format(i))
        ob = env.reset()
        done = False
        env_time = 0
        while not done:
            action = policy.act(ob)
            ob, _, done, info = env.step(action)
            event = info['event']
            if render:
                env.render('human')  # robocentric=True, save_to_file=True)
            env_time += env._get_dt()
            if isinstance(event, Danger):
                too_close += 1
                min_dist.append(event.min_dist)

        if isinstance(event, ReachGoal):
            success += 1
            success_times.append(env_time)
        elif isinstance(event, Collision):
            collision += 1
            collision_cases.append(i)
            collision_times.append(env_time)
        elif isinstance(event, CollisionOtherAgent):
            collision_other_agent += 1
            collision_other_agent_cases.append(i)
            collision_other_agent_times.append(env_time)
        elif isinstance(event, Timeout):
            timeout += 1
            timeout_cases.append(i)
            timeout_times.append(env_time)
        else:
            raise ValueError('Invalid end signal from environment')

    success_rate = success / float(num_episodes)
    collision_rate = collision / float(num_episodes)
    collision_other_agent_rate = collision_other_agent / \
        float(num_episodes)
    assert success + collision + timeout + collision_other_agent == num_episodes
    avg_nav_time = sum(success_times) / float(len(success_times)
                                              ) if success_times else np.nan

    print(
        """has success rate: {:.2f}, collision rate: {:.2f},
        collision from other agents rate: {:.2f}, nav time: {:.2f}, total reward: {:.4f}""".format(
            success_rate,
            collision_rate,
            collision_other_agent_rate,
            avg_nav_time,
            np.mean(cumulative_rewards)
        )
    )
    total_time = sum(success_times + collision_times + collision_other_agent_times + timeout_times)
    print(
        'Frequency of being in danger: %.2f and average min separate distance in danger: %.2f',
        too_close / float(total_time),
        np.mean(min_dist))

    if print_failure:
        print('Collision cases: ' + ' '.join([str(x) for x in collision_cases]))
        print('Collision from other agent cases: ' + ' '.join([str(x) for x in collision_other_agent_cases]))
        print('Timeout cases: ' + ' '.join([str(x) for x in timeout_cases]))

    return success_rate, avg_nav_time


if __name__ == '__main__':
    args, _ = parse_common_args()

    if args.environment is None or args.environment == "navreptrain":
        env = NavRepTrainEncodedEnv(args.backend, args.encoding, silent=True, scenario='test')
        policy = NavRepCPolicy()
    else:
        raise NotImplementedError

    run_test_episodes(env, policy, render=args.render)



import os
import time
import numpy as np
import pandas as pd
from matplotlib import pyplot as plt
from matplotlib import dates as md
import matplotlib
from rich.console import Console
from rich.table import Table
from datetime import timedelta

from navrep.tools.commonargs import parse_plotting_args

matplotlib.rcParams['mathtext.fontset'] = 'stix'
matplotlib.rcParams['font.family'] = 'STIXGeneral'

MILLION = 1000000
_N = 66


def moving_average(x, w):
    mva = np.convolve(x, np.ones(w), 'valid') / w
    if len(x) > len(mva):
        mva = np.pad(mva, (len(x) - len(mva), 0))
    if len(mva) > len(x):
        mva = mva[:len(x)]
    return mva

def smooth(x, weight):
    """ Weight between 0 and 1 """
    last = x[0]  # First value in the plot (first timestep)
    smoothed = list()
    for point in x:
        smoothed_val = last * weight + (1 - weight) * point  # Calculate smoothed value
        smoothed.append(smoothed_val)                        # Save it
        last = smoothed_val                                  # Anchor the last smoothed value
    return smoothed


SCENARIO_AXES = {
    "irosasl1"         : (0, 0),
    "irosasl2"         : (0, 1),
    "irosasl3"         : (0, 2),
    "irosasl4"         : (0, 3),
    "irosasl5"         : (0, 4),
    "irosasl6"         : (0, 5),
    "irosasl_office_j1": (1, 0),
    "irosasl_office_j2": (1, 1),
    "irosasl_office_j3": (1, 2),
    "irosasl_office_j4": (1, 3),
    "irosasl_office_j5": (1, 4),
    "irosasl_office_j6": (1, 5),
    "irosunity_scene_map1": (2, 0),
    "irosunity_scene_map2": (2, 1),
    "irosunity_scene_map3": (2, 2),
    "irosunity_scene_map4": (2, 3),
    "irosunity_scene_map5": (2, 4),
    "irosunity_scene_map6": (2, 5),
    "rlasl1"           : (3, 0),
    "rlasl2"           : (3, 1),
    "rlasl3"           : (3, 2),
    "rlasl_office_j1"  : (4, 0),
    "rlasl_office_j2"  : (4, 1),
    "rlasl_office_j3"  : (4, 2),
    "rlunity_scene_map1"  : (5, 0),
    "rlunity_scene_map2"  : (5, 1),
    "rlunity_scene_map3"  : (5, 2),
    "marktrain"        : (6, 0),
    "marksimple"       : (7, 0),
    "markcomplex"      : (7, 1),
    "marktm1"          : (7, 2),
    "marktm2"          : (7, 3),
    "marktm3"          : (7, 4),
}
ERROR_IF_MISSING = True

def parse_logfiles(navrep_dirs):
    best_navrep_names = [os.path.basename(path) for path in navrep_dirs]

    all_logpaths = []
    all_parents = []
    for name, dir_ in zip(best_navrep_names, navrep_dirs):
        logdir = os.path.join(dir_, "logs/gym")
        try:
            logfiles = sorted([file for file in os.listdir(logdir) if ".csv" in file])
        except FileNotFoundError:
            logfiles = []
        logpaths = [os.path.join(logdir, logfile) for logfile in logfiles]
        logparents = [name for _ in logfiles]
        all_logpaths.extend(logpaths)
        all_parents.extend(logparents)
    return all_logpaths, all_parents

def plot_per_scenario_training_progress(logdirs, scenario_axes, error_if_missing,
                                        x_axis="train_steps", paper_ready=False):
    SHOW_DIFFICULTY = True
    ANIMATE = False
    logpaths, parents = parse_logfiles(logdirs)

    # get set of all scenarios in all logpaths
    all_scenarios = []
    for logpath in logpaths:
        S = pd.read_csv(logpath)
        scenarios = sorted(list(set(S["scenario"].values)))
        all_scenarios.extend(scenarios)
    all_scenarios = sorted(list(set(all_scenarios)))

    print()
    print("Plotting scenario rewards")
    print()
    plt.ion()
    plt.figure("scenario rewards")
    plt.clf()
    N_ROWS, N_COLS = np.max(np.array(list(scenario_axes.values())), axis=0) + 1
    fig, axes = plt.subplots(N_ROWS, N_COLS, num="scenario rewards", sharex=True, sharey=True)
    axes = np.array(axes).reshape((N_ROWS, N_COLS))
    fig.suptitle("Per-Scenario Training Progress")
    # delete empty axes
    # not_present = [scenario for scenario in scenario_axes.keys() if scenario not in all_scenarios]
    # for scenario in not_present:
    #     ax_i, ax_j = scenario_axes[scenario]
    #     ax = axes[ax_i][ax_j]
    #     fig.delaxes(ax)

    console = Console()
    table = Table(show_header=True, header_style="bold magenta")
    table.add_column("Parent", style="dim")
    table.add_column("Name")
    table.add_column("Steps", justify="right")
    table.add_column("Reward", justify="right")

    lines = []
    legends = []
    total_train_time = 0.
    i_frame = 0
    for logpath, parent in zip(logpaths, parents):
        logname = os.path.basename(logpath)
        line = None
        color, style = get_color_and_style(logpath)
        S = pd.read_csv(logpath)
        add_train_steps_to_statistics(S)
        for scenario in all_scenarios:
            try:
                ax_i, ax_j = scenario_axes[scenario]
            except KeyError:
                if error_if_missing:
                    raise
                else:
                    continue
            ax = axes[ax_i][ax_j]
            scenario_S = S[S["scenario"] == scenario]
            if x_axis == "train_steps":
                x = scenario_S["train_steps"].values
                x = x / MILLION
            elif x_axis == "total_steps":
                x = scenario_S["total_steps"].values
                x = x / MILLION
            elif x_axis == "wall_time":
                try:
                    x = scenario_S["wall_time"].values
                except KeyError:
                    print("{} has no wall_time info".format(logpath))
                    continue
                x = md.epoch2num(x)
            else:
                raise NotImplementedError
            if len(x) == 0:
                continue
            ylabel = "reward"
            n = np.max(scenario_S["train_steps"].values) / MILLION
            rewards = scenario_S["reward"].values
#             x = np.concatenate([[0], x])
#             rewards = np.concatenate([[0], rewards])
            smooth_rewards = smooth(rewards, 0.99)
            if scenario == "navreptrain":
                y = smooth_rewards
                linestyle = "solid"
                # y as progress rate
                if True:
                    if x_axis == "wall_time":
                        y = scenario_S["train_steps"].values / MILLION
                        SHOW_DIFFICULTY = False
                        linestyle = style
                        ylabel = "Million train steps"
                        # predicted end
                        if n < _N:
                            predicted_end = (x[-1] - x[0]) * 1. * _N / n + x[0]
                            ax.plot([x[-1], predicted_end], [n, _N],
                                    linestyle=linestyle, color=color, linewidth=1., alpha=0.2)
                # plot main reward line
                tmp, = ax.plot(x, y, linewidth=1, linestyle=linestyle, color=color)
                color = tmp.get_c()
                # add line for difficulty
                if SHOW_DIFFICULTY:
#                 ax.plot(x, np.concatenate([[0], scenario_S["num_walls"].values]),
#                         linestyle="--", color=color, linewidth=1)
                    ax.plot(x, scenario_S["num_walls"].values,
                            linestyle="--", color=color, linewidth=1)
                # did the training stall?
                if x_axis == "wall_time":
                    if n < _N:
                        last = scenario_S["wall_time"].values[-1]
                        elapsed_hours_since_last_update = (time.time() - last) / 3600.
                        if elapsed_hours_since_last_update >= 6:
                            print("WARNING: {} may be stalled "
                                  "(incomplete and last update {} hours ago)".format(
                                      logpath, timedelta(hours=elapsed_hours_since_last_update)))
            elif scenario == "navrepval":
                y = rewards
                ylabel = "success rate [%]"
                # plot main reward line
                if ANIMATE:
                    yanim = np.ones_like(y) * np.nan
                    line, = ax.plot(x, yanim, linestyle=style, linewidth=1, color=color)
                    line.set_linewidth(2)
                    if x[-1] < 60:
                        continue
                    for i in range(len(y)):
                        yanim[i] = y[i]
                        line.set_ydata(yanim)
                        ax.set_xlim([-10, 66])
                        ax.set_ylim([-2, 100])
                        plt.pause(0.01)
                        plt.savefig("/tmp/plot_best_gym_{:05}.png".format(i_frame))
                        i_frame += 1
                        if x[i] > 66:
                            break
                    line.set_linewidth(1)
                else:
                    line, = ax.plot(x, y, linestyle=style, linewidth=1, color=color)
                color = line.get_c()
                top = np.max(y)
                # add best hline
                ax.axhline(top, color=color, linestyle=style, linewidth=1, alpha=0.5)
                # circle best score
                if not paper_ready:
                    ax.scatter(x[np.argmax(y)], top, marker='o', facecolor="none", edgecolor=color)
                # predicted end
                if x_axis == "wall_time":
                    train_time = x[-1] - x[0]
                    total_train_time += train_time
                    if n < _N:
                        predicted_end = train_time * _N * 1. / n + x[0]
                        ax.plot([x[-1], predicted_end], [y[-1], y[-1]],
                                linestyle=style, color=color, linewidth=1., alpha=0.2)
                        ax.scatter(predicted_end, y[-1], marker='x', facecolor=color)
                # table
                table.add_row(
                    parent,
                    logname,
                    "{:.1f}".format(n) if n > _N else "[dim]{:.1f}[/dim]".format(n),
                    "{:.1f}".format(top)
                )
            else:
                y = rewards
                # plot main reward line
                line, = ax.plot(x, smooth_rewards, linewidth=1, color=color)
                color = line.get_c()
                # add episode reward scatter
                ax.plot(x, y, color=color, marker=',', linewidth=0, label=scenario)
            # add vertical line at end of finished runs
            if x_axis == "wall_time":
                if n > _N:
                    ax.axvline(x[-1], linestyle=style, linewidth=1, color=color)
                else:
                    ax.scatter(x[-1], y[-1], marker='>', facecolor="none", edgecolor=color)
            ax.set_ylim([-1,101])
            ax.set_ylabel(ylabel)
            if x_axis == "train_steps":
                ax.set_xlabel("Million Train Steps")
            elif x_axis == "total_steps":
                ax.set_xlabel("Million Steps (Eval env - Train env is ~8x larger)")
            elif x_axis == "wall_time":
                ax.set_xlabel("Wall Time")
                xfmt = md.DateFormatter('%d-%b-%Y %H:%M:%S')
                ax.xaxis.set_major_formatter(xfmt)
                from matplotlib.ticker import MultipleLocator
                ax.xaxis.set_minor_locator(MultipleLocator(1))
                ax.xaxis.set_major_locator(MultipleLocator(3))
                ax.grid(which='minor', axis='x', linestyle='-')
                ax.grid(which='major', axis='x', linestyle='-')
            ax.set_title(scenario)
        if line is not None:
            lines.append(line)
            legends.append(parent + ": " + logname)
    # add lucia best hline
    try:
        ax_i, ax_j = scenario_axes["navrepval"]
    except KeyError:
        pass
    ax = axes[ax_i][ax_j]
    line = ax.axhline(85., color="red", linewidth=1, alpha=0.5)
    lines.append(line)
    legends.append("soadrl")
    # add current time
    if x_axis == "wall_time":
        print("Total train time: {} days".format(total_train_time))
        try:
            ax_i, ax_j = scenario_axes["navrepval"]
        except KeyError:
            pass
        ax = axes[ax_i][ax_j]
        ax.axvline(md.epoch2num(time.time()), color='k', linewidth=1)
        try:
            ax_i, ax_j = scenario_axes["navreptrain"]
        except KeyError:
            pass
        ax = axes[ax_i][ax_j]
        ax.axvline(md.epoch2num(time.time()), color='k', linewidth=1)

    console.print(table)
    if not paper_ready:
        fig.legend(lines, legends, bbox_to_anchor=(1.05, 1.))

def add_train_steps_to_statistics(S):
    # here we assume 10000 env.steps happened for every 20 episodes of eval env
    # but with 6 envs, that's 60000 training steps
    is_eval_scenario = S["scenario"] == "navreptrain"
    is_20th_eval_scenario = np.mod(np.cumsum(is_eval_scenario), 20) == 0
    train_steps = np.cumsum(is_20th_eval_scenario * 60000)
    S["train_steps"] = train_steps


backend_colors = {
    "VAE1D_LSTM": "lightskyblue",
    "VAE_LSTM": "cornflowerblue",
    "VAE1DLSTM": "mediumaquamarine",
    "VAELSTM": "mediumseagreen",
    "GPT1D": "khaki",
    "GPT": "gold",
    "E2E1D": "lightgrey",
    "E2E": "grey",
}
envname_colors = {
    "e2e1dnavreptrain": "lightgrey",
    "e2enavreptrain": "grey",
    "lucianavreptrain": "red",
    "guldenring": "plum",
}
encoding_styles = {
    "V_ONLY": "dotted",
    "M_ONLY": "dashed",
    "VM": "solid"
}

envname_labels = {
    "e2e1dnavreptrain": "End-to-end (1D)",
    "e2enavreptrain": "End-to-end",
    "lucianavreptrain": "SOADRL",
    "guldenring": "Guldenring et al.",
}
backend_labels = {
    "VAE1D_LSTM": "Modular (1D)",
    "VAE_LSTM": "Modular",
    "VAE1DLSTM": "Joint (1D)",
    "VAELSTM": "Joint",
    "GPT1D": "Transformer (1D)",
    "GPT": "Transformer",
    "E2E1D": "End-to-end (1D)",
    "E2E": "End-to-end",
}
encoding_labels = {
    "V_ONLY": "z only",
    "M_ONLY": "h only",
    "VM": "z+h"
}


def get_color_and_style(logpath):
    backend, encoding = get_backend_and_encoding(logpath)
    envname = get_envname(logpath)
    return color_and_style(encoding, backend, envname)

def color_and_style(encoding, backend, envname):
    color = None
    style = None
    if backend is not None:
        color = backend_colors[backend]
    if encoding is not None:
        style = encoding_styles[encoding]
    # if no backend, look if at least the envname is known
    if color is None:
        if envname in envname_colors:
            color = envname_colors[envname]
    return color, style

def get_backend_and_encoding(logpath):
    string = logpath.split("PPO_")[-1]
    backend = None
    encoding = None
    # find backend in string
    for k in backend_colors:
        if string.startswith(k):
            backend = k
            string = string.split(k+"_")[1]
            break
    # find encoding in remainder string
    for k in encoding_styles:
        if string.startswith(k):
            encoding = k
            break
    return backend, encoding

def get_envname(logpath):
    return os.path.basename(logpath).split("_")[0]

def get_label(envname, backend, encoding):
    label = None
    if envname in envname_labels:
        label = envname_labels[envname]
    else:
        try:
            label = "{}, {}".format(backend_labels[backend], encoding_labels[encoding])
        except KeyError:
            label = "{}, {}, {}".format(envname, backend, encoding)
    return label

def get_date(logpath):
    string = "_".join(os.path.basename(logpath).split("_")[1:])
    string = string.split("_PPO")[0]
    return string


if __name__ == "__main__":
    args, _ = parse_plotting_args()

    if args.scenario is not None:
        SCENARIO_AXES = {args.scenario : (0, 0)}  # show only 1 scenario
        ERROR_IF_MISSING = False
    if args.scenario == "navreptrain":
        SCENARIO_AXES = {args.scenario : (0, 0), "navrepval" : (1, 0)}
    if args.x_axis is None:
        args.x_axis = "train_steps"

    while True:
        logdir = args.logdir
        if logdir is None:
            logdir = "~/navrep"
        logdirs = [os.path.expanduser(logdir),]
        plot_per_scenario_training_progress(logdirs, SCENARIO_AXES, ERROR_IF_MISSING,
                                            x_axis=args.x_axis)
#         plot_per_scenario_training_progress(logdirs, SCENARIO_AXES, ERROR_IF_MISSING)
        plt.pause(60)



# This Python file uses the following encoding: utf-8
"""autogenerated by genpy from frame_msgs/PersonTrajectoryEntry.msg. Do not edit."""
import sys
python3 = True if sys.hexversion > 0x03000000 else False
import genpy
import struct

import geometry_msgs.msg
import genpy

class PersonTrajectoryEntry(genpy.Message):
  _md5sum = "116bbf92aab7b80413f5fd46c81faddc"
  _type = "frame_msgs/PersonTrajectoryEntry"
  _has_header = False #flag to mark the presence of a Header object
  _full_text = """# Message defining an entry of a person trajectory.
#

duration    age             # age of the track
bool        is_occluded     # if the track is currently not matched by a detection
uint64      detection_id    # id of the corresponding detection in the current cycle (undefined if occluded)

# The following fields are extracted from the Kalman state x and its covariance C

geometry_msgs/PoseWithCovariance    pose   # pose of the track (z value and orientation might not be set, check if corresponding variance on diagonal is > 99999)
geometry_msgs/TwistWithCovariance   twist  # velocity of the track (z value and rotational velocities might not be set, check if corresponding variance on diagonal is > 99999)

================================================================================
MSG: geometry_msgs/PoseWithCovariance
# This represents a pose in free space with uncertainty.

Pose pose

# Row-major representation of the 6x6 covariance matrix
# The orientation parameters use a fixed-axis representation.
# In order, the parameters are:
# (x, y, z, rotation about X axis, rotation about Y axis, rotation about Z axis)
float64[36] covariance

================================================================================
MSG: geometry_msgs/Pose
# A representation of pose in free space, composed of position and orientation. 
Point position
Quaternion orientation

================================================================================
MSG: geometry_msgs/Point
# This contains the position of a point in free space
float64 x
float64 y
float64 z

================================================================================
MSG: geometry_msgs/Quaternion
# This represents an orientation in free space in quaternion form.

float64 x
float64 y
float64 z
float64 w

================================================================================
MSG: geometry_msgs/TwistWithCovariance
# This expresses velocity in free space with uncertainty.

Twist twist

# Row-major representation of the 6x6 covariance matrix
# The orientation parameters use a fixed-axis representation.
# In order, the parameters are:
# (x, y, z, rotation about X axis, rotation about Y axis, rotation about Z axis)
float64[36] covariance

================================================================================
MSG: geometry_msgs/Twist
# This expresses velocity in free space broken into its linear and angular parts.
Vector3  linear
Vector3  angular

================================================================================
MSG: geometry_msgs/Vector3
# This represents a vector in free space. 
# It is only meant to represent a direction. Therefore, it does not
# make sense to apply a translation to it (e.g., when applying a 
# generic rigid transformation to a Vector3, tf2 will only apply the
# rotation). If you want your data to be translatable too, use the
# geometry_msgs/Point message instead.

float64 x
float64 y
float64 z"""
  __slots__ = ['age','is_occluded','detection_id','pose','twist']
  _slot_types = ['duration','bool','uint64','geometry_msgs/PoseWithCovariance','geometry_msgs/TwistWithCovariance']

  def __init__(self, *args, **kwds):
    """
    Constructor. Any message fields that are implicitly/explicitly
    set to None will be assigned a default value. The recommend
    use is keyword arguments as this is more robust to future message
    changes.  You cannot mix in-order arguments and keyword arguments.

    The available fields are:
       age,is_occluded,detection_id,pose,twist

    :param args: complete set of field values, in .msg order
    :param kwds: use keyword arguments corresponding to message field names
    to set specific fields.
    """
    if args or kwds:
      super(PersonTrajectoryEntry, self).__init__(*args, **kwds)
      #message fields cannot be None, assign default values for those that are
      if self.age is None:
        self.age = genpy.Duration()
      if self.is_occluded is None:
        self.is_occluded = False
      if self.detection_id is None:
        self.detection_id = 0
      if self.pose is None:
        self.pose = geometry_msgs.msg.PoseWithCovariance()
      if self.twist is None:
        self.twist = geometry_msgs.msg.TwistWithCovariance()
    else:
      self.age = genpy.Duration()
      self.is_occluded = False
      self.detection_id = 0
      self.pose = geometry_msgs.msg.PoseWithCovariance()
      self.twist = geometry_msgs.msg.TwistWithCovariance()

  def _get_types(self):
    """
    internal API method
    """
    return self._slot_types

  def serialize(self, buff):
    """
    serialize message into buffer
    :param buff: buffer, ``StringIO``
    """
    try:
      _x = self
      buff.write(_get_struct_2iBQ7d().pack(_x.age.secs, _x.age.nsecs, _x.is_occluded, _x.detection_id, _x.pose.pose.position.x, _x.pose.pose.position.y, _x.pose.pose.position.z, _x.pose.pose.orientation.x, _x.pose.pose.orientation.y, _x.pose.pose.orientation.z, _x.pose.pose.orientation.w))
      buff.write(_get_struct_36d().pack(*self.pose.covariance))
      _x = self
      buff.write(_get_struct_6d().pack(_x.twist.twist.linear.x, _x.twist.twist.linear.y, _x.twist.twist.linear.z, _x.twist.twist.angular.x, _x.twist.twist.angular.y, _x.twist.twist.angular.z))
      buff.write(_get_struct_36d().pack(*self.twist.covariance))
    except struct.error as se: self._check_types(struct.error("%s: '%s' when writing '%s'" % (type(se), str(se), str(locals().get('_x', self)))))
    except TypeError as te: self._check_types(ValueError("%s: '%s' when writing '%s'" % (type(te), str(te), str(locals().get('_x', self)))))

  def deserialize(self, str):
    """
    unpack serialized message in str into this message instance
    :param str: byte array of serialized message, ``str``
    """
    try:
      if self.age is None:
        self.age = genpy.Duration()
      if self.pose is None:
        self.pose = geometry_msgs.msg.PoseWithCovariance()
      if self.twist is None:
        self.twist = geometry_msgs.msg.TwistWithCovariance()
      end = 0
      _x = self
      start = end
      end += 73
      (_x.age.secs, _x.age.nsecs, _x.is_occluded, _x.detection_id, _x.pose.pose.position.x, _x.pose.pose.position.y, _x.pose.pose.position.z, _x.pose.pose.orientation.x, _x.pose.pose.orientation.y, _x.pose.pose.orientation.z, _x.pose.pose.orientation.w,) = _get_struct_2iBQ7d().unpack(str[start:end])
      self.is_occluded = bool(self.is_occluded)
      start = end
      end += 288
      self.pose.covariance = _get_struct_36d().unpack(str[start:end])
      _x = self
      start = end
      end += 48
      (_x.twist.twist.linear.x, _x.twist.twist.linear.y, _x.twist.twist.linear.z, _x.twist.twist.angular.x, _x.twist.twist.angular.y, _x.twist.twist.angular.z,) = _get_struct_6d().unpack(str[start:end])
      start = end
      end += 288
      self.twist.covariance = _get_struct_36d().unpack(str[start:end])
      self.age.canon()
      return self
    except struct.error as e:
      raise genpy.DeserializationError(e) #most likely buffer underfill


  def serialize_numpy(self, buff, numpy):
    """
    serialize message with numpy array types into buffer
    :param buff: buffer, ``StringIO``
    :param numpy: numpy python module
    """
    try:
      _x = self
      buff.write(_get_struct_2iBQ7d().pack(_x.age.secs, _x.age.nsecs, _x.is_occluded, _x.detection_id, _x.pose.pose.position.x, _x.pose.pose.position.y, _x.pose.pose.position.z, _x.pose.pose.orientation.x, _x.pose.pose.orientation.y, _x.pose.pose.orientation.z, _x.pose.pose.orientation.w))
      buff.write(self.pose.covariance.tostring())
      _x = self
      buff.write(_get_struct_6d().pack(_x.twist.twist.linear.x, _x.twist.twist.linear.y, _x.twist.twist.linear.z, _x.twist.twist.angular.x, _x.twist.twist.angular.y, _x.twist.twist.angular.z))
      buff.write(self.twist.covariance.tostring())
    except struct.error as se: self._check_types(struct.error("%s: '%s' when writing '%s'" % (type(se), str(se), str(locals().get('_x', self)))))
    except TypeError as te: self._check_types(ValueError("%s: '%s' when writing '%s'" % (type(te), str(te), str(locals().get('_x', self)))))

  def deserialize_numpy(self, str, numpy):
    """
    unpack serialized message in str into this message instance using numpy for array types
    :param str: byte array of serialized message, ``str``
    :param numpy: numpy python module
    """
    try:
      if self.age is None:
        self.age = genpy.Duration()
      if self.pose is None:
        self.pose = geometry_msgs.msg.PoseWithCovariance()
      if self.twist is None:
        self.twist = geometry_msgs.msg.TwistWithCovariance()
      end = 0
      _x = self
      start = end
      end += 73
      (_x.age.secs, _x.age.nsecs, _x.is_occluded, _x.detection_id, _x.pose.pose.position.x, _x.pose.pose.position.y, _x.pose.pose.position.z, _x.pose.pose.orientation.x, _x.pose.pose.orientation.y, _x.pose.pose.orientation.z, _x.pose.pose.orientation.w,) = _get_struct_2iBQ7d().unpack(str[start:end])
      self.is_occluded = bool(self.is_occluded)
      start = end
      end += 288
      self.pose.covariance = numpy.frombuffer(str[start:end], dtype=numpy.float64, count=36)
      _x = self
      start = end
      end += 48
      (_x.twist.twist.linear.x, _x.twist.twist.linear.y, _x.twist.twist.linear.z, _x.twist.twist.angular.x, _x.twist.twist.angular.y, _x.twist.twist.angular.z,) = _get_struct_6d().unpack(str[start:end])
      start = end
      end += 288
      self.twist.covariance = numpy.frombuffer(str[start:end], dtype=numpy.float64, count=36)
      self.age.canon()
      return self
    except struct.error as e:
      raise genpy.DeserializationError(e) #most likely buffer underfill

_struct_I = genpy.struct_I
def _get_struct_I():
    global _struct_I
    return _struct_I
_struct_36d = None
def _get_struct_36d():
    global _struct_36d
    if _struct_36d is None:
        _struct_36d = struct.Struct("<36d")
    return _struct_36d
_struct_2iBQ7d = None
def _get_struct_2iBQ7d():
    global _struct_2iBQ7d
    if _struct_2iBQ7d is None:
        _struct_2iBQ7d = struct.Struct("<2iBQ7d")
    return _struct_2iBQ7d
_struct_6d = None
def _get_struct_6d():
    global _struct_6d
    if _struct_6d is None:
        _struct_6d = struct.Struct("<6d")
    return _struct_6d



#!/usr/bin/env python
from navrep.envs.toytcndreamenv import ToyTCNDreamEnv
import numpy as np
from pyniel.python_tools.timetools import WalltimeRate

if __name__ == "__main__":
    # setting up the sim ---------------
    denv = ToyTCNDreamEnv(temperature=0.05)
    # run interactively ----------------------
    realtime_rate = WalltimeRate(1.0 / denv.DT)
    action = np.array([0.0, 0.0, 0.0])
    restart = [False]
    exit = [False]
    boost = [False]
    # key fetching
    from pyglet.window import key

    def key_press(k, mod):
        if k == 0xFF0D or k == key.ESCAPE:
            exit[0] = True
        if k == key.R:
            restart[0] = True
        if k in [key.RIGHT, key.D]:
            action[2] = -0.2
        if k in [key.LEFT, key.A]:
            action[2] = +0.2
        if k in [key.UP, key.W]:
            action[0] = +0.5
        if k in [key.DOWN, key.S]:
            action[0] = -0.3
        if k in [key.E]:
            action[1] = -0.4
        if k in [key.Q]:
            action[1] = +0.4
        if k in [key.LSHIFT]:
            boost[0] = True

    def key_release(k, mod):  # reverse action of pressed
        if k in [key.RIGHT, key.D] and action[2] == -0.2:
            action[2] = 0
        if k in [key.LEFT, key.A] and action[2] == +0.2:
            action[2] = 0
        if k in [key.UP, key.W] and action[0] == +0.5:
            action[0] = 0
        if k in [key.DOWN, key.S] and action[0] == -0.3:
            action[0] = 0
        if k in [key.E] and action[1] == -0.4:
            action[1] = 0
        if k in [key.Q] and action[1] == +0.4:
            action[1] = 0
        if k in [key.LSHIFT]:
            boost[0] = False

    denv.render()
    denv.viewer.window.on_key_press = key_press
    denv.viewer.window.on_key_release = key_release
    i = 0
    while not exit[0]:
        i += 1
        obs, rew, done, info = denv.step(action)
        denv.render()
        #         impglet
        #         pygage.get_buffer_manager().get_color_buffer().save("/tmp/denv{:05}.png".format(i))
        if done or restart[0]:
            print("Resetting")
            denv.reset()
            restart[0] = False
        if not boost[0]:
            realtime_rate.sleep()
denv.viewer.close()



import numpy as np
from timeit import default_timer as timer
from tqdm import tqdm

from navrep.envs.navreptrainenv import NavRepTrainEnv
from navrep.tools.commonargs import parse_common_args

if __name__ == "__main__":
    args, _ = parse_common_args()

    n = args.n
    if n is None:
        n = 1000000

    env = NavRepTrainEnv(scenario='train', silent=True, adaptive=False, collect_statistics=False)
    env.reset()

    action = np.array([0., 0., 0.])

    tic = timer()
    for i in tqdm(range(n)):
        env.step(action)
    toc = timer()
    elapsed = toc-tic

    print("Executed {} simulation steps in {:.1f} seconds.".format(n, elapsed))



import json
import numpy as np
import os
import tensorflow as tf

_64 = 64

def reset_graph():
    if "sess" in globals() and sess:
        sess.close()
    tf.reset_default_graph()


class ConvVAE(object):
    def __init__(
        self,
        z_size=32,
        batch_size=1,
        learning_rate=0.0001,
        kl_tolerance=0.5,
        is_training=False,
        reuse=False,
        channels=1,
    ):
        self.z_size = z_size
        self.batch_size = batch_size
        self.learning_rate = learning_rate
        self.is_training = is_training
        self.kl_tolerance = kl_tolerance
        self.reuse = reuse
        self.channels = channels
        with tf.variable_scope("conv_vae", reuse=self.reuse):
            self._build_graph()
        self._init_session()

    def _build_graph(self):
        self.g = tf.Graph()
        with self.g.as_default():

            self.x = tf.placeholder(tf.float32, shape=[None, _64, _64, self.channels])

            # Encoder
            h = tf.layers.conv2d(
                self.x, 32, 4, strides=2, activation=tf.nn.relu, name="enc_conv1"
            )
            h = tf.layers.conv2d(
                h, 64, 4, strides=2, activation=tf.nn.relu, name="enc_conv2"
            )
            h = tf.layers.conv2d(
                h, 128, 4, strides=2, activation=tf.nn.relu, name="enc_conv3"
            )
            h = tf.layers.conv2d(
                h, 256, 4, strides=2, activation=tf.nn.relu, name="enc_conv4"
            )
            h = tf.reshape(h, [-1, 2 * 2 * 256])

            # VAE
            self.mu = tf.layers.dense(h, self.z_size, name="enc_fc_mu")
            self.logvar = tf.layers.dense(h, self.z_size, name="enc_fc_log_var")
            self.sigma = tf.exp(self.logvar / 2.0)
            self.epsilon = tf.random_normal([self.batch_size, self.z_size])
            if self.is_training:
                self.z = self.mu + self.sigma * self.epsilon
            else:
                self.z = self.mu

            # Decoder
            h = tf.layers.dense(self.z, 4 * 256, name="dec_fc")
            h = tf.reshape(h, [-1, 1, 1, 4 * 256])
            h = tf.layers.conv2d_transpose(
                h, 128, 5, strides=2, activation=tf.nn.relu, name="dec_deconv1"
            )
            h = tf.layers.conv2d_transpose(
                h, 64, 5, strides=2, activation=tf.nn.relu, name="dec_deconv2"
            )
            h = tf.layers.conv2d_transpose(
                h, 32, 6, strides=2, activation=tf.nn.relu, name="dec_deconv3"
            )
            self.y = tf.layers.conv2d_transpose(
                h,
                self.channels,
                6,
                strides=2,
                activation=tf.nn.sigmoid,
                name="dec_deconv4",
            )

            # train ops
            if self.is_training:
                self.global_step = tf.Variable(0, name="global_step", trainable=False)

                eps = 1e-6  # avoid taking log of zero

                # reconstruction loss
                self.r_loss = tf.reduce_sum(
                    tf.square(self.x - self.y), reduction_indices=[1, 2, 3]
                )
                self.r_loss = tf.reduce_mean(self.r_loss)

                # augmented kl loss per dim
                self.kl_loss = -0.5 * tf.reduce_sum(
                    (1 + self.logvar - tf.square(self.mu) - tf.exp(self.logvar)),
                    reduction_indices=1,
                )
                self.kl_loss = tf.maximum(self.kl_loss, self.kl_tolerance * self.z_size)
                self.kl_loss = tf.reduce_mean(self.kl_loss)

                self.loss = self.r_loss + self.kl_loss

                # training
                self.lr = tf.Variable(self.learning_rate, trainable=False)
                self.optimizer = tf.train.AdamOptimizer(self.lr)
                grads = self.optimizer.compute_gradients(
                    self.loss
                )  # can potentially clip gradients here.

                self.train_op = self.optimizer.apply_gradients(
                    grads, global_step=self.global_step, name="train_step"
                )

            # initialize vars
            self.init = tf.global_variables_initializer()

            t_vars = tf.trainable_variables()
            self.assign_ops = {}
            for var in t_vars:
                # if var.name.startswith('conv_vae'):
                pshape = var.get_shape()
                pl = tf.placeholder(tf.float32, pshape, var.name[:-2] + "_placeholder")
                assign_op = var.assign(pl)
                self.assign_ops[var] = (assign_op, pl)

    def _init_session(self):
        """Launch TensorFlow session and initialize variables"""
        self.sess = tf.Session(graph=self.g)
        self.sess.run(self.init)

    def close_sess(self):
        """ Close TensorFlow session """
        self.sess.close()

    def encode(self, x):
        return self.sess.run(self.z, feed_dict={self.x: x})

    def encode_mu_logvar(self, x):
        (mu, logvar) = self.sess.run([self.mu, self.logvar], feed_dict={self.x: x})
        return mu, logvar

    def decode(self, z):
        return self.sess.run(self.y, feed_dict={self.z: z})

    def encode_decode(self, x):
        return self.sess.run(self.y, feed_dict={self.x: x})

    def get_model_params(self):
        # get trainable params.
        model_names = []
        model_params = []
        model_shapes = []
        with self.g.as_default():
            t_vars = tf.trainable_variables()
            for var in t_vars:
                # if var.name.startswith('conv_vae'):
                param_name = var.name
                p = self.sess.run(var)
                model_names.append(param_name)
                params = np.round(p * 10000).astype(np.int).tolist()
                model_params.append(params)
                model_shapes.append(p.shape)
        return model_params, model_shapes, model_names

    def print_trainable_params(self):
        trainable_params = 0
        with self.g.as_default():
            t_vars = tf.trainable_variables()
            for var in t_vars:
                trainable_params += np.prod(var.get_shape().as_list())
        print("VAE Trainable parameters: {}".format(trainable_params))

    def get_random_model_params(self, stdev=0.5):
        # get random params.
        _, mshape, _ = self.get_model_params()
        rparam = []
        for s in mshape:
            # rparam.append(np.random.randn(*s)*stdev)
            rparam.append(np.random.standard_cauchy(s) * stdev)  # spice things up
        return rparam

    def set_model_params(self, params):
        with self.g.as_default():
            t_vars = tf.trainable_variables()
            idx = 0
            for var in t_vars:
                # if var.name.startswith('conv_vae'):
                pshape = tuple(var.get_shape().as_list())
                p = np.array(params[idx])
                assert pshape == p.shape, "inconsistent shape"
                assign_op, pl = self.assign_ops[var]
                self.sess.run(assign_op, feed_dict={pl.name: p / 10000.0})
                idx += 1

    def load_json(self, jsonfile="vae.json"):
        with open(jsonfile, "r") as f:
            params = json.load(f)
        self.set_model_params(params)

    def save_json(self, jsonfile="vae.json"):
        model_params, model_shapes, model_names = self.get_model_params()
        qparams = []
        for p in model_params:
            qparams.append(p)
        with open(jsonfile, "wt") as outfile:
            json.dump(
                qparams, outfile, sort_keys=True, indent=0, separators=(",", ": ")
            )

    def set_random_params(self, stdev=0.5):
        rparam = self.get_random_model_params(stdev)
        self.set_model_params(rparam)

    def save_model(self, model_save_path):
        sess = self.sess
        with self.g.as_default():
            saver = tf.train.Saver(tf.global_variables())
        checkpoint_path = os.path.join(model_save_path, "vae")
        tf.logging.info("saving model %s.", checkpoint_path)
        saver.save(sess, checkpoint_path, 0)  # just keep one

    def load_checkpoint(self, checkpoint_path):
        sess = self.sess
        with self.g.as_default():
            saver = tf.train.Saver(tf.global_variables())
        ckpt = tf.train.get_checkpoint_state(checkpoint_path)
        print("loading model", ckpt.model_checkpoint_path)
        tf.logging.info("Loading model %s.", ckpt.model_checkpoint_path)
        saver.restore(sess, ckpt.model_checkpoint_path)



"""
Transformer blocks taken from Karpathy's minGPT (MIT License) https://github.com/karpathy/minGPT
"""

import random
import numpy as np
import math
import logging
import torch
import torch.nn as nn
from torch.nn import functional as F

from navrep.models.torchvae import VAE

logger = logging.getLogger(__name__)

_A = 3
_G = 2

STATE_NORM_FACTOR = 25.  # maximum typical goal distance, meters

def set_seed(seed):
    random.seed(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)
    torch.cuda.manual_seed_all(seed)

class GPTConfig:
    embd_pdrop = 0.1
    resid_pdrop = 0.1
    attn_pdrop = 0.1

    n_layer = 8
    n_head = 8
    n_action = _A  # forward v, side v, rot v
    n_states = _G  # goal_x, goal_y (in robot frame)
    image_channels = 1

    def __init__(self, block_size, n_embd):
        self.block_size = block_size
        self.n_embd = n_embd  # good size is 64

class CausalSelfAttention(nn.Module):
    def __init__(self, config):
        super().__init__()
        assert config.n_embd % config.n_head == 0
        # key, query, value projections for all heads
        self.key = nn.Linear(config.n_embd, config.n_embd)
        self.query = nn.Linear(config.n_embd, config.n_embd)
        self.value = nn.Linear(config.n_embd, config.n_embd)
        # regularization
        self.attn_drop = nn.Dropout(config.attn_pdrop)
        self.resid_drop = nn.Dropout(config.resid_pdrop)
        # output projection
        self.proj = nn.Linear(config.n_embd, config.n_embd)
        # causal mask to ensure that attention is only applied to the left in the input sequence
        self.register_buffer(
            "mask",
            torch.tril(torch.ones(config.block_size, config.block_size)).view(
                1, 1, config.block_size, config.block_size
            ),
        )
        self.n_head = config.n_head

    def forward(self, x, layer_past=None):
        B, T, C = x.size()

        # calculate query, key, values for all heads in batch and move head forward to be the batch dim
        k = (
            self.key(x).view(B, T, self.n_head, C // self.n_head).transpose(1, 2)
        )  # (B, nh, T, hs)
        q = (
            self.query(x).view(B, T, self.n_head, C // self.n_head).transpose(1, 2)
        )  # (B, nh, T, hs)
        v = (
            self.value(x).view(B, T, self.n_head, C // self.n_head).transpose(1, 2)
        )  # (B, nh, T, hs)

        # causal self-attention; Self-attend: (B, nh, T, hs) x (B, nh, hs, T) -> (B, nh, T, T)
        att = (q @ k.transpose(-2, -1)) * (1.0 / math.sqrt(k.size(-1)))
        att = att.masked_fill(
            self.mask[:, :, :T, :T] == 0, -1e10
        )  # todo: just use float('-inf') instead?
        att = F.softmax(att, dim=-1)
        att = self.attn_drop(att)
        y = att @ v  # (B, nh, T, T) x (B, nh, T, hs) -> (B, nh, T, hs)
        y = (
            y.transpose(1, 2).contiguous().view(B, T, C)
        )  # re-assemble all head outputs side by side

        # output projection
        y = self.resid_drop(self.proj(y))
        return y

class Block(nn.Module):

    def __init__(self, config):
        super().__init__()
        self.ln1 = nn.LayerNorm(config.n_embd)
        self.ln2 = nn.LayerNorm(config.n_embd)
        self.attn = CausalSelfAttention(config)
        self.mlp = nn.Sequential(
            nn.Linear(config.n_embd, 4 * config.n_embd),
            nn.GELU(),
            nn.Linear(4 * config.n_embd, config.n_embd),
            nn.Dropout(config.resid_pdrop),
        )

    def forward(self, x):
        x = x + self.attn(self.ln1(x))
        x = x + self.mlp(self.ln2(x))
        return x

class GPT(nn.Module):
    """ A prediction model based on openAI's GPT architecture """

    def __init__(self, config, gpu=True):
        super().__init__()

        # input embedding stem
        self.convVAE = VAE(z_dim=config.n_embd, gpu=gpu, image_channels=config.image_channels)
        self.action_emb = nn.Linear(config.n_action, config.n_embd)
        self.pos_emb = nn.Parameter(torch.zeros(1, config.block_size, config.n_embd))
        self.state_emb = nn.Linear(config.n_states, config.n_embd)
        self.drop = nn.Dropout(config.embd_pdrop)
        # transformer
        self.blocks = nn.Sequential(*[Block(config) for _ in range(config.n_layer)])
        # decoder head
        self.ln_f = nn.LayerNorm(config.n_embd)
        self.z_head = nn.Linear(config.n_embd, config.n_embd)
        self.state_head = nn.Linear(config.n_embd, config.n_states)

        self.block_size = config.block_size
        self.n_embd = config.n_embd
        self.apply(self._init_weights)

        self.gpu = gpu

        logger.info(
            "number of parameters: %e", sum(p.numel() for p in self.parameters())
        )

    def _init_weights(self, module):
        if isinstance(module, (nn.Linear, nn.Embedding)):
            module.weight.data.normal_(mean=0.0, std=0.02)
            if isinstance(module, nn.Linear) and module.bias is not None:
                module.bias.data.zero_()
        elif isinstance(module, nn.LayerNorm):
            module.bias.data.zero_()
            module.weight.data.fill_(1.0)

    def get_block_size(self):
        return self.block_size

    def forward(self, img, state, action, dones, targets=None, h=None):
        """
        img: (batch, sequence, CH, W, H) [0, 1]
        action: (batch, sequence, A) [-inf, inf]
        state: (batch, sequence, S) [-inf, inf]
        dones:  (batch, sequence,) {0, 1}
        targets: None or (img_targets, state_targets)
            img_targets: same shape as img
            state_targets: same shape as state

        OUTPUTS
        img_pred: same shape as img
        state_pred: same shape as state
        loss: torch loss
        """
        b, t, CH, W, H = img.size()
        _, _, A = action.size()
        _, _, S = state.size()
        assert t <= self.block_size, "Cannot forward, model block size is exhausted."

        # encode embedding with vae
        z, mu, logvar = self.convVAE.encode(img.view(b * t, CH, W, H))  # each image maps to a vector
        token_embeddings = z.view(b, t, self.n_embd)
        state_embeddings = self.state_emb(state.view(b * t, S)).view(b, t, self.n_embd)
        action_embeddings = self.action_emb(action.view(b * t, A)).view(b, t, self.n_embd)
        position_embeddings = self.pos_emb[:, :t, :]  # each position maps to a (learnable) vector
        # forward the GPT model
        x = self.drop(token_embeddings + position_embeddings + action_embeddings + state_embeddings)
        x = self.blocks(x)
        x = self.ln_f(x)
        # store worldmodel embedding
        if h is not None:
            h[0] = x
        # decode embedding with vae
        z_pred = self.z_head(x.view(b * t, self.n_embd))
        img_rec = self.convVAE.decode(z).view(b, t, CH, W, H)
        img_pred = self.convVAE.decode(z_pred).view(b, t, CH, W, H)
        state_pred = self.state_head(x.view(b * t, self.n_embd)).view(b, t, S)

        # if we are given some desired targets also calculate the loss
        loss = None
        if targets is not None:
            img_targets, state_targets = targets
            rec_loss = F.binary_cross_entropy(img_rec, img)  # input-reconstruction loss
            pred_loss = F.binary_cross_entropy(img_pred, img_targets)  # reconstructed prediction loss
            state_loss = F.mse_loss(state_pred, state_targets) / STATE_NORM_FACTOR**2
            KLD = -0.5 * torch.mean(1 + logvar - mu.pow(2) - logvar.exp())  # kullback leibler
            kld_tolerance = 0.5
            kld_weight = 0.001
            KLD = torch.max(KLD, kld_tolerance * torch.ones_like(KLD))
            loss = rec_loss + kld_weight * KLD + pred_loss + state_loss

        return img_pred, state_pred, loss

    def _to_correct_device(self, tensor):
        if self.gpu:
            if torch.cuda.is_available():
                device = torch.cuda.current_device()
                return tensor.to(device)
            else:
                print("WARNING: model created with gpu enabled, but no gpu found")
        return tensor

    def encode(self, img):
        """
        img: numpy (batch, W, H, CH)

        OUTPUTS
        z: (batch, Z)
        """
        b, W, H, CH = img.shape

        img_t = torch.tensor(np.moveaxis(img, -1, 1), dtype=torch.float)
        img_t = self._to_correct_device(img_t)

        z, mu, logvar = self.convVAE.encode(img_t)
        return z.detach().cpu().numpy()

    def encode_mu_logvar(self, img):
        """
        img: numpy (batch, W, H, CH)


        OUTPUTS
        mu: (batch, Z)
        logvar: (batch, Z)
        """
        b, W, H, CH = img.shape

        img_t = torch.tensor(np.moveaxis(img, -1, 1), dtype=torch.float)
        img_t = self._to_correct_device(img_t)

        z, mu, logvar = self.convVAE.encode(img_t)
        mu = mu.detach().cpu().numpy()
        logvar = logvar.detach().cpu().numpy()
        return mu, logvar

    def decode(self, z):
        """
        z: numpy (batch, Z)

        OUTPUTS
        img_rec: (batch, W, H, CH)
        """
        b, Z = z.shape

        z_t = torch.tensor(z, dtype=torch.float)
        z_t = self._to_correct_device(z_t)

        img_rec_t = self.convVAE.decode(z_t) # b, CH, W, H
        img_rec = np.moveaxis(img_rec_t.detach().cpu().numpy(), 1, -1)
        return img_rec

    def get_h(self, gpt_sequence):
        """ for compat with encodedenv
        gpt sequence is a list of dicts, one for each step in the sequence.
        each dict has
        "obs": numpy image (W, H, CH) [0, 1]
        "state": numpy (2,) [-inf, inf]
        "action": numpy (3,) [-inf, inf]
        """
        _b = 1  # batch size
        img = np.array([d["obs"] for d in gpt_sequence])  # t, W, H, CH
        img = np.moveaxis(img, -1, 1)
        img = img.reshape((_b, *img.shape))
        img_t = torch.tensor(img, dtype=torch.float)
        img_t = self._to_correct_device(img_t)
        state = np.array([d["state"] for d in gpt_sequence])  # t, 2
        state = state.reshape((_b, *state.shape))
        state_t = torch.tensor(state, dtype=torch.float)
        state_t = self._to_correct_device(state_t)
        action = np.array([d["action"] for d in gpt_sequence])  # t, 3
        action = action.reshape((_b, *action.shape))
        action_t = torch.tensor(action, dtype=torch.float)
        action_t = self._to_correct_device(action_t)
        dones = np.zeros((_b, len(gpt_sequence), 1))
        dones_t = torch.tensor(dones, dtype=torch.float)
        dones_t = self._to_correct_device(dones_t)
        h_container = [None]
        self.forward(img_t, state_t, action_t, dones_t, h=h_container)
        h = h_container[0].detach().cpu().numpy()
        h = h[0, -1]  # only batch, last item in sequence
        return h

    def get_next(self, gpt_sequence):
        """ for compat with encodedenv
        gpt sequence is a list of dicts, one for each step in the sequence.
        each dict has
        "obs": numpy image (W, H, CH) [0, 1]
        "state": numpy (2,) [-inf, inf]
        "action": numpy (3,) [-inf, inf]
        output:
        img_pred: (W, H, CH)
        state_pred: (2,)
        """
        _b = 1  # batch size
        img = np.array([d["obs"] for d in gpt_sequence])  # t, W, H, CH
        img = np.moveaxis(img, -1, 1)
        img = img.reshape((_b, *img.shape))
        img_t = torch.tensor(img, dtype=torch.float)
        img_t = self._to_correct_device(img_t)
        state = np.array([d["state"] for d in gpt_sequence])  # t, 2
        state = state.reshape((_b, *state.shape))
        state_t = torch.tensor(state, dtype=torch.float)
        state_t = self._to_correct_device(state_t)
        action = np.array([d["action"] for d in gpt_sequence])  # t, 3
        action = action.reshape((_b, *action.shape))
        action_t = torch.tensor(action, dtype=torch.float)
        action_t = self._to_correct_device(action_t)
        dones = np.zeros((_b, len(gpt_sequence), 1))
        dones_t = torch.tensor(dones, dtype=torch.float)
        dones_t = self._to_correct_device(dones_t)
        h_container = [None]
        img_pred_t, state_pred_t, _ = self.forward(img_t, state_t, action_t, dones_t, h=h_container)
        img_pred = img_pred_t.detach().cpu().numpy()
        img_pred = img_pred[0, -1]  # only batch, last item in sequence
        img_pred = np.moveaxis(img_pred, 0, -1)
        state_pred = state_pred_t.detach().cpu().numpy()
        state_pred = state_pred[0, -1]  # only batch, last item in sequence
        return img_pred, state_pred

def save_checkpoint(model, ckpt_path):
    if ckpt_path is not None:
        ckpt_model = model.module if hasattr(model, "module") else model
        logger.info("saving %s", ckpt_path)
        torch.save(ckpt_model.state_dict(), ckpt_path)

def load_checkpoint(model, ckpt_path, gpu=False):
    if gpu:
        device = torch.cuda.current_device()
        map_location = "cuda:{}".format(device)
    else:
        map_location = torch.device('cpu')
    state_dict = torch.load(ckpt_path, map_location=map_location)
    model.load_state_dict(state_dict)
    if gpu:
        model.to(device)



import torch
import torch.nn as nn

class Flatten(nn.Module):
    def forward(self, input):
        return input.reshape(input.size(0), -1)

class UnFlatten(nn.Module):
    def __init__(self, channels):
        super(UnFlatten, self).__init__()
        self.channels = channels

    def forward(self, input):
        return input.view(input.size(0), self.channels, 1, 1)

class UnFlatten1D(nn.Module):
    def __init__(self, channels, size):
        super(UnFlatten1D, self).__init__()
        self.channels = channels
        self.size = size

    def forward(self, input):
        return input.view(input.size(0), self.channels, self.size)

class VAE(nn.Module):
    def __init__(self, image_channels=1, h_dim=1024, z_dim=32, gpu=True):
        self.gpu = gpu
        super(VAE, self).__init__()
        self.encoder = nn.Sequential(
            nn.Conv2d(image_channels, 32, kernel_size=4, stride=2),
            nn.ReLU(),
            nn.Conv2d(32, 64, kernel_size=4, stride=2),
            nn.ReLU(),
            nn.Conv2d(64, 128, kernel_size=4, stride=2),
            nn.ReLU(),
            nn.Conv2d(128, 256, kernel_size=4, stride=2),
            nn.ReLU(),
            Flatten(),
        )

        self.fc1 = nn.Linear(h_dim, z_dim)
        self.fc2 = nn.Linear(h_dim, z_dim)
        self.fc3 = nn.Linear(z_dim, h_dim)

        self.decoder = nn.Sequential(
            UnFlatten(h_dim),
            nn.ConvTranspose2d(h_dim, 128, kernel_size=5, stride=2),
            nn.ReLU(),
            nn.ConvTranspose2d(128, 64, kernel_size=5, stride=2),
            nn.ReLU(),
            nn.ConvTranspose2d(64, 32, kernel_size=6, stride=2),
            nn.ReLU(),
            nn.ConvTranspose2d(32, image_channels, kernel_size=6, stride=2),
            nn.Sigmoid(),
        )

    def reparameterize(self, mu, logvar):
        std = logvar.mul(0.5).exp_()
        # return torch.normal(mu, std)
        if self.gpu:
            eps = torch.cuda.FloatTensor(*mu.size()).normal_()
        else:
            eps = torch.FloatTensor(*mu.size()).normal_()
        z = mu + std * eps
        return z

    def bottleneck(self, h):
        mu, logvar = self.fc1(h), self.fc2(h)
        z = self.reparameterize(mu, logvar)
        return z, mu, logvar

    def encode(self, x):
        h = self.encoder(x)
        z, mu, logvar = self.bottleneck(h)
        return z, mu, logvar

    def decode(self, z):
        z = self.fc3(z)
        z = self.decoder(z)
        return z

    def forward(self, x):
        z, mu, logvar = self.encode(x)
        z = self.decode(z)
        return z, mu, logvar

class VAE1D(VAE):
    def __init__(self, image_channels=1, h_dim=1024, z_dim=32, gpu=True):
        assert h_dim % 4 == 0

        super(VAE1D, self).__init__()
        self.gpu = gpu

        self.encoder = nn.Sequential(
            nn.Conv1d(image_channels, 32, kernel_size=8, stride=4),
            nn.ReLU(),
            nn.Conv1d(32, 64, kernel_size=9, stride=4),
            nn.ReLU(),
            nn.Conv1d(64, 128, kernel_size=6, stride=4),
            nn.ReLU(),
            nn.Conv1d(128, 256, kernel_size=4, stride=4),
            nn.ReLU(),
            Flatten(),
        )

        self.fc1 = nn.Linear(h_dim, z_dim)
        self.fc2 = nn.Linear(h_dim, z_dim)
        self.fc3 = nn.Linear(z_dim, h_dim)

        self.decoder = nn.Sequential(
            UnFlatten1D(h_dim//4, 4),  # (batch, 1024) -> (batch, 256, 4)
            nn.ConvTranspose1d(h_dim//4, 128, kernel_size=4, stride=4),
            nn.ReLU(),
            nn.ConvTranspose1d(128, 64, kernel_size=6, stride=4),
            nn.ReLU(),
            nn.ConvTranspose1d(64, 32, kernel_size=9, stride=4),
            nn.ReLU(),
            nn.ConvTranspose1d(32, image_channels, kernel_size=8, stride=4),
            nn.Sigmoid(),
        )



from navrep.envs.navreptrainenv import NavRepTrainEnv

class CADRLEnv(NavRepTrainEnv):
    def __init__(self, silent=False, legacy_mode=False, n_humans=10):
        super(CADRLEnv, self).__init__(
            scenario='train', silent=silent, legacy_mode=legacy_mode, adaptive=False, lidar_legs=False)
        self.soadrl_sim.num_circles = 0
        self.soadrl_sim.num_walls = 0
        self.soadrl_sim.human_num = n_humans

    def _add_border_obstacle(self):
        pass


if __name__ == "__main__":
    from navrep.tools.envplayer import EnvPlayer
    env = CADRLEnv()
    player = EnvPlayer(env)
